use strict;
use warnings;
use File::Spec;
use File::Copy::Recursive ('fcopy', 'rcopy', 'dircopy');

my $buildno = '0.2.2015.08.08';
my $devnull = File::Spec->devnull();

# options
my $numthreads = 1;
my $minident1 = 0.97;
my $minident2;
my $minident3;
my $minovllen = 100;
my $strand;
my $alnwiggle = 11;
my $assemblemode = 'minimus';
my $mappingmode = 'nucmer';
my $linkagemode = 'complete';
my $mergemode = 'normal';
my $minnseq;
my $maxnsplit;
my $procdepth = 0;
my $usecontigmembers;
my $addniter = 0;
my $firstaddniter;
my $replaceprecontig = 0;
my $rawreads;
my $noeid;
my $nodel;

# input/output
my @inputfiles;
my $outputbank;

# commands
my $assams = "\"$0\"";
my $toAmos_new;
my $hashoverlap;
my $ungappedoverlap;
my $tigger;
my $makeconsensus;
my $recallConsensus;
my $bank2fastq;
my $bank2fasta;
my $listReadPlacedStatus;
my $dumpreads;
my $bankreport;
my $banktransact;
my $bankcombine;
my $bankunlock;
my $pblat;
my $blat2nucmer;
my $nucmerAnnotate;
my $nucmer2ovl;
my $ovl2OVL;
my $nucmer;
my $casmlayout;

# global variables
my %nseqs;
my $format;
my @firstinputfiles;
my $converged;

# file handles
my $filehandleinput1;
my $filehandleinput2;
my $filehandleoutput1;
my $filehandleoutput2;
my $pipehandleinput1;
my $pipehandleinput2;
my $pipehandleoutput1;
my $pipehandleoutput2;

&main();

sub main {
	# print startup messages
	&printStartupMessage();
	# get command line arguments
	&getOptions();
	# check variable consistency
	&checkVariables();
	# check running environment
	&checkEnvironment();
	# recognize input format
	&recognizeFormat();
	# split sequences into several groups
	&splitSequences();
	# assemble splitted sequences
	&assembleSplittedSequences();
	# progressive merging of multiple assembly
	&mergeAssembly();
	exit(0);
}

sub printStartupMessage {
	print(STDERR <<"_END");
Assams $buildno
=======================================================================

This is a script to assemble many long sequences in parallel. Official
web site of this script is
http://www.fifthdimension.jp/products/assams/ .
To know script details, see above URL.

Copyright (C) 2011-2015  Akifumi S. Tanabe

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

_END
	# display usage if command line options were not specified
	unless (@ARGV) {
		&helpMessage();
	}
}

sub getOptions {
	# get arguments
	$outputbank = $ARGV[-1];
	if (-e $outputbank) {
		&errorMessage(__LINE__, "\"$outputbank\" already exists.");
	}
	{
		my @temp = glob("$outputbank*");
		if (@temp) {
			&errorMessage(__LINE__, "Temporary file already exists.");
		}
	}
	{
		my %inputfiles;
		for (my $i = 0; $i < scalar(@ARGV) - 1; $i ++) {
			if ($ARGV[$i] =~ /^-+assemble(?:mode)?=(exact|blat|pblat|blatfast|pblatfast|minimus|e|b|p|bf|pf|m)$/i) {
				if ($1 =~ /^(?:blat|pblat|b|p)$/i) {
					$assemblemode = 'blat';
				}
				elsif ($1 =~ /^(?:blatfast|pblatfast|bf|pf)$/i) {
					$assemblemode = 'blatfast';
				}
				elsif ($1 =~ /^(?:minimus|m)$/i) {
					$assemblemode = 'minimus';
				}
			}
			elsif ($ARGV[$i] =~ /^-+mapping(?:mode)?=(nucmer|none)$/i) {
				if ($1 =~ /^nucmer$/i) {
					$mappingmode = 'nucmer';
				}
				elsif ($1 =~ /^none$/i) {
					$mappingmode = 'none';
				}
			}
			elsif ($ARGV[$i] =~ /^-+rawreads?=(.+)$/i) {
				$rawreads = $1;
			}
			elsif ($ARGV[$i] =~ /^-+linkage(?:mode)?=(complete|single|c|s)$/i) {
				if ($1 =~ /^(?:complete|c)$/i) {
					$linkagemode = 'complete';
				}
				elsif ($1 =~ /^(?:single|s)$/i) {
					$linkagemode = 'single';
				}
			}
			elsif ($ARGV[$i] =~ /^-+merge(?:mode)?=(rough|lazy|draft|rapid|fast|middle|moderate|normal|accurate|slow|strict)$/i) {
				if ($1 =~ /^(?:accurate|slow|strict)$/i) {
					$mergemode = 'accurate';
				}
				elsif ($1 =~ /^(?:middle|moderate|normal)$/i) {
					$mergemode = 'normal';
				}
				elsif ($1 =~ /^(?:draft|rapid|fast)$/i) {
					$mergemode = 'draft';
				}
				elsif ($1 =~ /^(?:rough|lazy)$/i) {
					$mergemode = 'rough';
				}
			}
			elsif ($ARGV[$i] =~ /^-+(?:min(?:imum)?ident(?:ity|ities)?|m)=(.+)$/i) {
				my @minident = split(/,/, $1);
				if (scalar(@minident) > 3) {
					&errorMessage(__LINE__, "Too many minident were given.");
				}
				$minident1 = $minident[0];
				if ($minident[1]) {
					$minident2 = $minident[1];
				}
				if ($minident[2]) {
					$minident3 = $minident[2];
				}
			}
			elsif ($ARGV[$i] =~ /^-+min(?:imum)?(?:overlap|ovl)(?:length|len)=(\d+)$/i) {
				$minovllen = $1;
			}
			elsif ($ARGV[$i] =~ /^-+(?:strand|s)=(forward|plus|single|both|double)$/i) {
				if ($1 =~ /^(?:forward|plus|single)$/i) {
					$strand = ' -s';
				}
			}
			elsif ($ARGV[$i] =~ /^-+(?:(?:aln|align|alignment)wiggle|a)=(\d+)$/i) {
				$alnwiggle = $1;
			}
			elsif ($ARGV[$i] =~ /^-+(?:n|n(?:um)?threads?)=(\d+)$/i) {
				$numthreads = $1;
			}
			elsif ($ARGV[$i] =~ /^-+min(?:imum)?n(?:um)?seq(?:uence)?s?=(\d+)$/i) {
				$minnseq = $1;
			}
			elsif ($ARGV[$i] =~ /^-+max(?:imum)?n(?:um)?splits?=(\d+)$/i) {
				$maxnsplit = $1;
			}
			elsif ($ARGV[$i] =~ /^-+proc(?:ess)?depth=(\d+)$/i) {
				$procdepth = $1;
			}
			elsif ($ARGV[$i] =~ /^-+usecontigmembers$/i) {
				$usecontigmembers = 1;
			}
			elsif ($ARGV[$i] =~ /^-+addn(?:um)?iter(?:ation)?=(\d+)$/i) {
				$addniter = $1;
			}
			elsif ($ARGV[$i] =~ /^-+noeid$/i) {
				$noeid = 1;
			}
			elsif ($ARGV[$i] =~ /^-+nodel$/i) {
				$nodel = 1;
			}
			else {
				my @temp = glob($ARGV[$i]);
				if (scalar(@temp > 0)) {
					foreach (@temp) {
						$inputfiles{$_} = 1;
					}
				}
				else {
					&errorMessage(__LINE__, "Input file does not exist.");
				}
			}
		}
		@inputfiles = sort({$a cmp $b} keys(%inputfiles));
	}
}

sub checkVariables {
	unless (@inputfiles) {
		&errorMessage(__LINE__, "There is no input files.");
	}
	if (!defined($minident2)) {
		if ($mappingmode eq 'none' || $linkagemode eq 'single') {
			$minident2 = $minident1;
		}
		else {
			$minident2 = 1 - ((1 - $minident1) * 1.1);
		}
	}
	if (!defined($minident3)) {
		$minident3 = $minident1;
	}
	if ($minident1 != 0 && ($minident1 > 1 || $minident1 < 0.7)) {
		&errorMessage(__LINE__, "Minimum identity threshold is invalid.");
	}
	if ($minident2 != 0 && ($minident2 > 1 || $minident2 < 0.5)) {
		&errorMessage(__LINE__, "Minimum identity threshold is invalid.");
	}
	if ($minident3 != 0 && ($minident3 > 1 || $minident3 < 0.7)) {
		&errorMessage(__LINE__, "Minimum identity threshold is invalid.");
	}
	if ($minident1 > $minident2 && ($mergemode eq 'draft' || $mergemode eq 'rough')) {
		&errorMessage(__LINE__, "Fast merging is incompatible to larger minimum identity threshold 1 than minimum identity threshold 2.");
	}
	if ($minovllen > 0 && $minovllen < 50) {
		&errorMessage(__LINE__, "Minimum overlap length is too small.");
	}
	if ($alnwiggle > 100 || $alnwiggle < 2) {
		&errorMessage(__LINE__, "Alignment wiggle value is invalid.");
	}
	if ($numthreads < 1) {
		&errorMessage(__LINE__, "The number of threads is too small.");
	}
	if (!defined($minnseq)) {
		if ($assemblemode eq 'minimus') {
			$minnseq = 1000;
		}
		else {
			$minnseq = 5000;
		}
	}
	elsif ($minnseq < 2) {
		&errorMessage(__LINE__, "The minimum number of sequences is too small.");
	}
	if (defined($maxnsplit) && $maxnsplit < 1) {
		&errorMessage(__LINE__, "The maximum number of splits is too small.");
	}
	if ($procdepth > 1) {
		&errorMessage(__LINE__, "The process depth is too large.");
	}
	if ($usecontigmembers && $mappingmode ne 'none') {
		&errorMessage(__LINE__, "Mapping mode \"$mappingmode\" is incompatible to usecontigmembers\.");
	}
	if ($rawreads && $mappingmode eq 'none') {
		&errorMessage(__LINE__, "Mapping mode \"$mappingmode\" is incompatible to rawreads\.");
	}
	$firstaddniter = $addniter;
}

sub checkEnvironment {
	my $pathto;
	if ($ENV{'ASSAMSHOME'}) {
		$pathto = $ENV{'ASSAMSHOME'};
	}
	else {
		my $temp;
		if (-e '.assams') {
			$temp = '.assams';
		}
		elsif (-e $ENV{'HOME'} . '/.assams') {
			$temp = $ENV{'HOME'} . '/.assams';
		}
		elsif (-e '/etc/assams/.assams') {
			$temp = '/etc/assams/.assams';
		}
		if ($temp) {
			unless (open($filehandleinput1, "< $temp")) {
				&errorMessage(__LINE__, "Cannot read \"$temp\".");
			}
			while (<$filehandleinput1>) {
				if (/^\s*ASSAMSHOME\s*=\s*(\S[^\r\n]*)/) {
					$pathto = $1;
					$pathto =~ s/\s+$//;
					last;
				}
			}
			close($filehandleinput1);
		}
	}
	if ($pathto) {
		$pathto =~ s/^"(.+)"$/$1/;
		$pathto =~ s/\/$//;
		$pathto .= '/bin';
		if (!-e $pathto) {
			&errorMessage(__LINE__, "Cannot find \"$pathto\".");
		}
		$toAmos_new = "\"$pathto/toAmos_new\"";
		$hashoverlap = "\"$pathto/hash-overlap\"";
		$ungappedoverlap = "\"$pathto/ungapped-overlap\"";
		$tigger = "\"$pathto/tigger\"";
		$makeconsensus = "\"$pathto/make-consensus\"";
		$recallConsensus = "\"$pathto/recallConsensus\"";
		$bank2fastq = "\"$pathto/bank2fastq\"";
		$bank2fasta = "\"$pathto/bank2fasta\"";
		$listReadPlacedStatus = "\"$pathto/listReadPlacedStatus\"";
		$dumpreads = "\"$pathto/dumpreads\"";
		$bankreport = "\"$pathto/bank-report\"";
		$banktransact = "\"$pathto/bank-transact\"";
		$bankcombine = "\"$pathto/bank-combine\"";
		$bankunlock = "\"$pathto/bank-unlock\"";
		$pblat = "\"$pathto/pblat\"";
		$blat2nucmer = "\"$pathto/blat2nucmer\"";
		$nucmerAnnotate = "\"$pathto/nucmerAnnotate\"";
		$nucmer2ovl = "\"$pathto/nucmer2ovl\"";
		$ovl2OVL = "\"$pathto/ovl2OVL\"";
		$nucmer = "$pathto/nucmer";
		$casmlayout = "$pathto/casm-layout";
	}
	else {
		$toAmos_new = 'toAmos_new';
		$hashoverlap = 'hash-overlap';
		$ungappedoverlap = 'ungapped-overlap';
		$tigger = 'tigger';
		$makeconsensus = 'make-consensus';
		$recallConsensus = 'recallConsensus';
		$bank2fastq = 'bank2fastq';
		$bank2fasta = 'bank2fasta';
		$listReadPlacedStatus = 'listReadPlacedStatus';
		$dumpreads = 'dumpreads';
		$bankreport = 'bank-report';
		$banktransact = 'bank-transact';
		$bankcombine = 'bank-combine';
		$bankunlock = 'bank-unlock';
		$pblat = 'pblat';
		$blat2nucmer = 'blat2nucmer';
		$nucmerAnnotate = 'nucmerAnnotate';
		$nucmer2ovl = 'nucmer2ovl';
		$ovl2OVL = 'ovl2OVL';
		$nucmer = 'nucmer';
		$casmlayout = 'casm-layout';
	}
}

sub recognizeFormat {
	foreach my $inputfile (@inputfiles) {
		my @temp = glob("$inputfile/*");
		if (@temp && (!defined($format) || $format eq 'BANK')) {
			$format = 'BANK';
		}
		elsif (($inputfile =~ /\.afg$/i || $inputfile =~ /\.afg\.gz$/i) && (!defined($format) || $format eq 'AFG')) {
			$format = 'AFG';
		}
		elsif (!defined($format) || $format eq 'FASTQ') {
			$format = 'FASTQ';
		}
		else {
			&errorMessage(__LINE__, "Input formats are not homogeneous.\nFile: $inputfile");
		}
	}
}

sub splitSequences {
	if ($format eq 'FASTQ') {
		if ($usecontigmembers) {
			&errorMessage(__LINE__, "Cannot use contigmembers with FASTQ input files\.");
		}
		print(STDERR "Splitting sequences...");
		my $nseq = 0;
		# check the number of sequences
		for (my $i = 0; $i < scalar(@inputfiles); $i ++) {
			$filehandleinput1 = &readFile($inputfiles[$i]);
			while (<$filehandleinput1>) {
				if (/^.+\r?\n?$/) {
					$nseq ++;
				}
			}
			close($filehandleinput1);
		}
		if ($nseq % 4 == 0) {
			$nseq /= 4;
		}
		else {
			&errorMessage(__LINE__, "The input files are invalid.");
		}
		# calculate nseqs of each file
		my @nseqperfile;
		if ($nseq < $minnseq * 2) {
			push(@nseqperfile, $nseq);
		}
		elsif ($maxnsplit && int($nseq / $minnseq) > $maxnsplit) {
			my $temp = int($nseq / $maxnsplit);
			for (my $i = 0; $i < $maxnsplit; $i ++) {
				if ($i < $nseq % $maxnsplit) {
					push(@nseqperfile, $temp + 1);
				}
				else {
					push(@nseqperfile, $temp);
				}
			}
		}
		else {
			my $nsplit = int($nseq / $minnseq);
			my $temp = int($nseq / $nsplit);
			for (my $i = 0; $i < $nsplit; $i ++) {
				if ($i < $nseq % $nsplit) {
					push(@nseqperfile, $temp + 1);
				}
				else {
					push(@nseqperfile, $temp);
				}
			}
		}
		# make splitted files
		{
			my $tempnfile = 0;
			my $tempnseq = 0;
			my $seqname;
			my %tempseqs;
			my %tempqual;
			my %templen;
			for (my $i = 0; $i < scalar(@inputfiles); $i ++) {
				$filehandleinput1 = &readFile($inputfiles[$i]);
				my $tempnline = 1;
				while (<$filehandleinput1>) {
					s/\r?\n?$//;
					if ($tempnline % 4 == 1 && /^\@(.+)/) {
						$seqname = $1;
					}
					elsif ($tempnline % 4 == 2) {
						$tempseqs{$seqname} = $_;
						$templen{$seqname} = length($_);
					}
					elsif ($tempnline % 4 == 0) {
						$tempqual{$seqname} = $_;
						$tempnseq ++;
						if ($tempnseq == $nseqperfile[$tempnfile]) {
							unless (open($filehandleoutput1, "> $outputbank.$tempnfile.fastq")) {
								&errorMessage(__LINE__, "Cannot write \"$outputbank.$tempnfile.fastq\".");
							}
							foreach my $tempname (sort({$templen{$b} <=> $templen{$a} || $a cmp $b} keys(%templen))) {
								print($filehandleoutput1 "\@$tempname\n$tempseqs{$tempname}\n+\n$tempqual{$tempname}\n");
							}
							close($filehandleoutput1);
							undef(%tempseqs);
							undef(%tempqual);
							undef(%templen);
							$tempnseq = 0;
							$tempnfile ++;
						}
						undef($seqname);
					}
					$tempnline ++;
				}
				close($filehandleinput1);
			}
		}
		# make AMOS bank
		undef(@inputfiles);
		for (my $i = 0; $i < scalar(@nseqperfile); $i ++) {
			&FASTQtoBank("$outputbank.$i.fastq", "$outputbank.$i.bnk", 1);
			push(@inputfiles, "$outputbank.$i.bnk");
			$nseqs{"$outputbank.$i.bnk"} = $nseqperfile[$i];
		}
		print(STDERR "done.\n\n");
	}
	elsif ($format eq 'AFG' || $format eq 'BANK') {
		if ($format eq 'AFG') {
			my @tempinputfiles;
			for (my $i = 0; $i < scalar(@inputfiles); $i ++) {
				if ($usecontigmembers) {
					my $contigmembers = $inputfiles[$i];
					$contigmembers =~ s/\.afg$/\.contigmembers\.gz/i;
					$contigmembers =~ s/\.afg\.gz$/\.contigmembers\.gz/i;
					if (-e $contigmembers) {
						unless (fcopy($contigmembers, "$outputbank.$i.contigmembers.gz")) {
							&errorMessage(__LINE__, "Cannot copy \"$contigmembers\" to \"$outputbank.$i.contigmembers.gz\"\.");
						}
					}
					else {
						&errorMessage(__LINE__, "\"$contigmembers\" does not exist\.");
					}
				}
				if ($inputfiles[$i] =~ /\.gz$/i) {
					if (system("gzip -dc $inputfiles[$i] | $banktransact -c -f -b $outputbank.$i.bnk -m - 1> $devnull 2> $devnull")) {
						&errorMessage(__LINE__, "Cannot run \"gzip -dc $inputfiles[$i] | $banktransact -c -f -b $outputbank.$i.bnk -m -\".");
					}
				}
				else {
					if (system("$banktransact -c -f -b $outputbank.$i.bnk -m $inputfiles[$i] 1> $devnull 2> $devnull")) {
						&errorMessage(__LINE__, "Cannot run \"$banktransact -c -f -b $outputbank.$i.bnk -m $inputfiles[$i]\".");
					}
				}
				push(@tempinputfiles, "$outputbank.$i.bnk");
			}
			@inputfiles = @tempinputfiles;
		}
		for (my $i = 0; $i < scalar(@inputfiles); $i ++) {
			my $nseq = 0;
			unless (open($pipehandleinput1, "$dumpreads -r $inputfiles[$i] 2> $devnull |")) {
				&errorMessage(__LINE__, "Cannot run \"$dumpreads -r $inputfiles[$i]\".");
			}
			while (<$pipehandleinput1>) {
				if (/^>/) {
					$nseq ++;
				}
			}
			close($pipehandleinput1);
			if ($?) {
				&errorMessage(__LINE__, "Cannot run \"$dumpreads -r $inputfiles[$i]\".");
			}
			$nseqs{$inputfiles[$i]} = $nseq;
		}
	}
}

sub assembleSplittedSequences {
	# assemble sequences by minimus pipeline in parallel
	print(STDERR "Assembling sequences...\n");
	for (my $i = 0; $i < scalar(@inputfiles); $i ++) {
		if ($nseqs{$inputfiles[$i]} > 1 && !-e "$inputfiles[$i]/CTG.ifo") {
			print(STDERR "Processing $inputfiles[$i]...\n");
			#if ($linkagemode eq 'complete') {
			#	&assembleBank($inputfiles[$i], $minident1, 1, 0, 0);
			#}
			#else {
			&assembleBank($inputfiles[$i], $minident1, 0, 0, 0);
			#}
		}
	}
	print(STDERR "done.\n\n");
}

sub mergeAssembly {
	if ($addniter < 1 && scalar(@inputfiles) == 1 && $minident1 == $minident2) {
		print(STDERR "Generating output bank...\n");
		if ($format eq 'BANK') {
			unless (dircopy($inputfiles[0], "$outputbank.0.bnk")) {
				&errorMessage(__LINE__, "Cannot copy \"$inputfiles[0]\" to \"$outputbank.0.bnk\".");
			}
			$inputfiles[0] = "$outputbank.0.bnk";
		}
	}
	else {
		# progressive merge or iterative assemble process
		if (scalar(@inputfiles) == 1 || $addniter > 0) {
			print(STDERR "Running iterative assemble...\n");
		}
		else {
			print(STDERR "Merging assemblies progressively...\n");
		}
		@firstinputfiles = @inputfiles;
		my $step = 1;
		while (scalar(@inputfiles) > 1 || $step == 1 || ($addniter > 0 && !$converged)) {
			if (scalar(@inputfiles) > 1) {
				print(STDERR "Step $step...\n");
			}
			# determine merge combination
			if (scalar(@inputfiles) > 2) {
				# count the number of contigs+singletons
				my %ncontigs;
				for (my $i = 0; $i < scalar(@inputfiles); $i ++) {
					$ncontigs{$inputfiles[$i]} = &countAllContigsSingletonsinBank($inputfiles[$i]);
				}
				my @temp = sort({$ncontigs{$a} <=> $ncontigs{$b} || $a cmp $b} keys(%ncontigs));
				undef(@inputfiles);
				# set Nth large and Nth small banks to the pair
				while (@temp) {
					if (scalar(@temp == 1)) {
						push(@inputfiles, shift(@temp));
					}
					else {
						push(@inputfiles, pop(@temp), shift(@temp));
					}
				}
			}
			my @currentinputfiles = @inputfiles;
			my $noutput = 1;
			my @nextinputfiles;
			while (@currentinputfiles) {
				my @focalinputfiles;
				push(@focalinputfiles, shift(@currentinputfiles));
				my $prefix = $outputbank . '_step' . $step . '_' . $noutput;
				# store second bank if available
				if (@currentinputfiles) {
					push(@focalinputfiles, shift(@currentinputfiles));
				}
				# print focal input files
				if ($addniter > 0 && scalar(@inputfiles) == 1) {
					print(STDERR "Assembling $focalinputfiles[0] iteratively...\n");
				}
				elsif (scalar(@focalinputfiles) == 2) {
					print(STDERR "Merging $focalinputfiles[0] and $focalinputfiles[1]...\n");
				}
				if (scalar(@inputfiles) > 1 && scalar(@focalinputfiles) == 1) {
					my $contigmembers = $focalinputfiles[0];
					$contigmembers =~ s/\.bnk$/.contigmembers.gz/;
					if ($step > 1) {
						unless (rename($focalinputfiles[0], "$prefix.bnk")) {
							&errorMessage(__LINE__, "Cannot rename \"$focalinputfiles[0]\" to \"$prefix.bnk\".");
						}
						if (-e $contigmembers) {
							unless (rename($contigmembers, "$prefix.contigmembers.gz")) {
								&errorMessage(__LINE__, "Cannot rename \"$contigmembers\" to \"$prefix.contigmembers.gz\".");
							}
						}
					}
					else {
						unless (dircopy($focalinputfiles[0], "$prefix.bnk")) {
							&errorMessage(__LINE__, "Cannot copy \"$focalinputfiles[0]\" to \"$prefix.bnk\".");
						}
						if (-e $contigmembers) {
							unless (fcopy($contigmembers, "$prefix.contigmembers.gz")) {
								&errorMessage(__LINE__, "Cannot copy \"$contigmembers\" to \"$prefix.contigmembers.gz\".");
							}
						}
					}
				}
				else {
					# make combined bank
					my $nseqfirst = 0;
					my $prevncontigs = 0;
					if ($mergemode eq 'rough' && scalar(@inputfiles) > 1 || $mergemode eq 'draft' && scalar(@inputfiles) > 2 || $mergemode eq 'normal' && $step > 1 && scalar(@inputfiles) > 2) {
						foreach my $focalinputfile (@focalinputfiles) {
							&dumpAllContigsSingletonsfromBank($focalinputfile, "$focalinputfile.fastq", 0);
							unless (open($filehandleinput1, "< $focalinputfile.fastq")) {
								&errorMessage(__LINE__, "Cannot read \"$focalinputfile.fastq\".");
							}
							my $tempnline = 1;
							unless (open($filehandleoutput1, ">> $prefix.fastq")) {
								&errorMessage(__LINE__, "Cannot write \"$prefix.fastq\".");
							}
							while (<$filehandleinput1>) {
								if ($tempnline % 4 == 1 && scalar(@inputfiles) == 1 && $addniter > 0 && $procdepth == 0) {
									s/^\@contig_/\@precontig_/;
								}
								print($filehandleoutput1 $_);
								$tempnline ++;
							}
							close($filehandleoutput1);
							close($filehandleinput1);
							unless (unlink("$focalinputfile.fastq")) {
								&errorMessage(__LINE__, "Cannot delete \"$focalinputfile.fastq\".");
							}
							if (!$nseqfirst && ($mergemode eq 'rough' || $mergemode eq 'draft' && scalar(@inputfiles) > 2 || $mergemode eq 'normal' && $step > 1 && scalar(@inputfiles) > 2)) {
								$nseqfirst = int($tempnline / 4);
							}
							$prevncontigs += int($tempnline / 4);
						}
					}
					else {
						my %tempseqs;
						my %tempqual;
						my %templen;
						foreach my $focalinputfile (@focalinputfiles) {
							&dumpAllContigsSingletonsfromBank($focalinputfile, "$focalinputfile.fastq", 0);
							unless (open($filehandleinput1, "< $focalinputfile.fastq")) {
								&errorMessage(__LINE__, "Cannot read \"$focalinputfile.fastq\".");
							}
							my $tempnline = 1;
							my $seqname;
							while (<$filehandleinput1>) {
								s/\r?\n?$//;
								if ($tempnline % 4 == 1 && /^\@(.+)/) {
									$seqname = $1;
									if (scalar(@inputfiles) == 1 && $addniter > 0 && $procdepth == 0) {
										$seqname =~ s/^contig_/precontig_/;
									}
								}
								elsif ($tempnline % 4 == 2) {
									$tempseqs{$seqname} = $_;
									$templen{$seqname} = length($_);
								}
								elsif ($tempnline % 4 == 0) {
									$tempqual{$seqname} = $_;
									undef($seqname);
								}
								$tempnline ++;
							}
							close($filehandleinput1);
							unless (unlink("$focalinputfile.fastq")) {
								&errorMessage(__LINE__, "Cannot delete \"$focalinputfile.fastq\".");
							}
							$prevncontigs += int($tempnline / 4);
						}
						unless (open($filehandleoutput1, ">> $prefix.fastq")) {
							&errorMessage(__LINE__, "Cannot write \"$prefix.fastq\".");
						}
						if (scalar(@inputfiles) == 1 && $addniter > 0 && $procdepth == 0) {
							my @tempname = sort({$templen{$b} <=> $templen{$a} || $a cmp $b} keys(%templen));
							while (@tempname) {
								my @tempname2;
								push(@tempname2, shift(@tempname));
								if (@tempname) {
									push(@tempname2, pop(@tempname));
								}
								foreach my $tempname (@tempname2) {
									print($filehandleoutput1 "\@$tempname\n$tempseqs{$tempname}\n+\n$tempqual{$tempname}\n");
								}
							}
						}
						else {
							foreach my $tempname (sort({$templen{$b} <=> $templen{$a} || $a cmp $b} keys(%templen))) {
								print($filehandleoutput1 "\@$tempname\n$tempseqs{$tempname}\n+\n$tempqual{$tempname}\n");
							}
						}
						close($filehandleoutput1);
					}
					# secondary assemble
					if (scalar(@inputfiles) == 1 && $addniter > 0 && $procdepth == 0 && $prevncontigs > 1) {
						&reassembleFASTQ("$prefix.fastq", "$prefix.temp1.bnk", $minident2, 1, 1, 'single', 'none');
					}
					else {
						&FASTQtoBank("$prefix.fastq", "$prefix.temp1.bnk", 1);
						if ($prevncontigs > 1) {
							#if ($linkagemode eq 'complete') {
							#	&assembleBank("$prefix.temp1.bnk", $minident2, 1, 0, $nseqfirst);
							#}
							#else {
							&assembleBank("$prefix.temp1.bnk", $minident2, 0, 0, $nseqfirst);
							#}
						}
					}
					if (scalar(@inputfiles) == 1 && $addniter > 0 && $procdepth == 0) {
						$replaceprecontig = 1;
						$addniter --;
						if ($prevncontigs == &countAllContigsSingletonsinBank("$prefix.temp1.bnk")) {
							$converged = 1;
						}
					}
					# not last step
					if (scalar(@inputfiles) > 2 || $mappingmode eq 'none' || ($addniter > 0 && !$converged)) {
						# write contig - raw reads association
						&associateReadstoContig("$prefix.temp1.bnk", "$prefix.bnk", @focalinputfiles);
						unlink("$prefix.temp1.contigmembers.gz");
					}
					# last step
					else {
						print(STDERR "Mapping raw reads to supercontigs...\n");
						# map raw reads to secondary contig
						&mapRawreadstoSecondarycontig("$prefix.temp1.bnk", "$prefix.bnk", @focalinputfiles);
						if (!$nodel) {
							unlink("$prefix.temp1.contigmembers.gz");
						}
					}
				}
				push(@nextinputfiles, "$prefix.bnk");
				$noutput ++;
			}
			if (scalar(@inputfiles) > 2 || ($addniter > 0 && !$converged)) {
				print(STDERR "Preparing next step...\n");
			}
			else {
				print(STDERR "Finishing last step...\n");
			}
			# delete temporary bank
			if ($step > 1) {
				if (!$nodel) {
					&deleteBanks(@inputfiles);
					&deleteContigMembers(@inputfiles);
				}
			}
			# renew inputfiles
			@inputfiles = @nextinputfiles;
			$step ++;
		}
		# delete temporary bank
		if ($format eq 'FASTQ' || $format eq 'AFG') {
			if (!$nodel) {
				&deleteBanks(@firstinputfiles);
				&deleteContigMembers(@firstinputfiles);
			}
		}
	}
	# add contig eid
	unless ($noeid) {
		&addEIDstoContigs($inputfiles[0]);
	}
	# write contig - raw reads association
	&makeContigMembers($inputfiles[0]);
	# rename contig members
	{
		my $contigmembers = $inputfiles[0];
		$contigmembers =~ s/\.bnk$/.contigmembers.gz/;
		my $outputcontigmembers = $outputbank;
		$outputcontigmembers =~ s/\.bnk$/.contigmembers.gz/;
		$outputcontigmembers =~ s/\.fastq$/.contigmembers.gz/i;
		$outputcontigmembers =~ s/\.fastq\.gz$/.contigmembers.gz/i;
		$outputcontigmembers =~ s/\.afg$/.contigmembers.gz/i;
		$outputcontigmembers =~ s/\.afg\.gz$/.contigmembers.gz/i;
		if ($outputcontigmembers !~ /\.contigmembers\.gz/i) {
			$outputcontigmembers .= '.contigmembers.gz';
		}
		if (-e $contigmembers) {
			unless (rename($contigmembers, $outputcontigmembers)) {
				&errorMessage(__LINE__, "Cannot rename \"$contigmembers\" to \"$outputcontigmembers\".");
			}
		}
		else {
			&errorMessage(__LINE__, "\"$contigmembers\" does not exist.");
		}
	}
	# make final output
	if ($outputbank =~ /\.fastq$/i) {
		&dumpAllContigsSingletonsfromBank($inputfiles[0], $outputbank, 1);
	}
	elsif ($outputbank =~ /\.fastq.gz$/i) {
		my $tempoutput = $outputbank;
		$tempoutput =~ s/\.gz$//i;
		&dumpAllContigsSingletonsfromBank($inputfiles[0], $tempoutput, 1);
		if (system("gzip $tempoutput")) {
			&errorMessage(__LINE__, "Cannot run \"gzip $tempoutput\".");
		}
	}
	elsif ($outputbank =~ /\.afg$/i) {
		if (system("$bankreport -p -b $inputfiles[0] 2> $devnull > $outputbank")) {
			&errorMessage(__LINE__, "Cannot run \"$bankreport -p -b $inputfiles[0] > $outputbank\".");
		}
	}
	elsif ($outputbank =~ /\.afg\.gz$/i) {
		if (system("$bankreport -p -b $inputfiles[0] 2> $devnull | gzip -c > $outputbank")) {
			&errorMessage(__LINE__, "Cannot run \"$bankreport -p -b $inputfiles[0] | gzip -c > $outputbank\".");
		}
	}
	elsif ($outputbank =~ /\.bnk$/i) {
		unless (rename($inputfiles[0], $outputbank)) {
			&errorMessage(__LINE__, "Cannot rename \"$inputfiles[0]\" to \"$outputbank\".");
		}
	}
	else {
		my $tempoutput = $outputbank;
		$tempoutput =~ s/\.gz$//i;
		&dumpAllContigsSingletonsfromBank($inputfiles[0], "$tempoutput.fastq", 1);
		if (system("gzip $tempoutput.fastq")) {
			&errorMessage(__LINE__, "Cannot run \"gzip $tempoutput.fastq\".");
		}
		if (system("$bankreport -p -b $inputfiles[0] 2> $devnull | gzip -c > $tempoutput.afg.gz")) {
			&errorMessage(__LINE__, "Cannot run \"$bankreport -p -b $inputfiles[0] | gzip -c > $tempoutput.afg.gz\".");
		}
		unless (rename($inputfiles[0], "$tempoutput.bnk")) {
			&errorMessage(__LINE__, "Cannot rename \"$inputfiles[0]\" to \"$tempoutput.bnk\".");
		}
	}
	if (!$nodel && -e $inputfiles[0]) {
		&deleteBanks($inputfiles[0]);
		&deleteContigMembers($inputfiles[0]);
	}
	if (!$nodel) {
		&deleteBanks(glob("$outputbank*.temp"));
	}
	print(STDERR "done.\n\n");
}

sub mapRawreadstoSecondarycontig {
	my $banknamein = shift(@_);
	my $banknameout = shift(@_);
	my @primarybank = @_;
	my %association;
	my %singletons;
	{
		# read association
		my %primarycontigs;
		my ($secondarycontigmembers, $secondarysingletons) = &getContigMembersfromBank($banknamein);
		my ($primarycontigmembers, $primarysingletons) = &getContigMembersfromBank(@primarybank);
		if ($secondarycontigmembers->{$banknamein}) {
			foreach my $contigname (keys(%{$secondarycontigmembers->{$banknamein}})) {
				foreach my $member (@{$secondarycontigmembers->{$banknamein}->{$contigname}}) {
					if ($member =~ /^contig_(.+)_(\d+)$/ || ($replaceprecontig && $member =~ /^precontig_(.+)_(\d+)$/)) {
						foreach my $rawmember (@{$primarycontigmembers->{$1}->{"contig_$2"}}) {
							$association{$rawmember} = $contigname;
						}
					}
					else {
						$association{$member} = $contigname;
					}
				}
			}
		}
		if ($secondarysingletons->{$banknamein}) {
			foreach my $member (@{$secondarysingletons->{$banknamein}}) {
				if ($member =~ /^contig_(.+)_(\d+)$/ || ($replaceprecontig && $member =~ /^precontig_(.+)_(\d+)$/)) {
					$primarycontigs{$member} = 1;
					foreach my $rawmember (@{$primarycontigmembers->{$1}->{"contig_$2"}}) {
						$association{$rawmember} = $member;
					}
				}
				else {
					$singletons{$member} = 1;
				}
			}
		}
		# export contig sequence to FASTA file
		my $numcontig = 1;
		if ($secondarycontigmembers->{$banknamein}) {
			foreach my $contigname (keys(%{$secondarycontigmembers->{$banknamein}})) {
				$contigname =~ /^contig_(\d+)$/;
				my $contigiid = $1;
				if ($contigiid > $numcontig) {
					$numcontig = $contigiid;
				}
				&dumpContigsfromBankAsFASTA($banknamein, "$banknameout.$contigname.fasta", $contigiid);
				if (!-e "$banknameout.$contigname.fasta" || -z "$banknameout.$contigname.fasta") {
					&dumpReadsfromBankAsFASTA($banknamein, "$banknameout.$contigname.fasta", $contigname, 1);
					if (!-e "$banknameout.$contigname.fasta" || -z "$banknameout.$contigname.fasta") {
						&errorMessage(__LINE__, "Cannot dump \"$contigname\" from \"$banknamein\" to \"$banknameout.$contigname.fasta\".");
					}
				}
			}
			$numcontig ++;
		}
		foreach my $primarycontig (sort({$a cmp $b} keys(%primarycontigs))) {
			$primarycontig =~ /contig_(.+)_(\d+)$/;
			my $primarybank = $1;
			my $contigiid = $2;
			&dumpContigsfromBankAsFASTA($primarybank, "$banknameout.contig_$numcontig.fasta", $contigiid);
			if (!-e "$banknameout.contig_$numcontig.fasta" || -z "$banknameout.contig_$numcontig.fasta") {
				&dumpReadsfromBankAsFASTA($primarybank, "$banknameout.contig_$numcontig.fasta", "contig_$contigiid", 1);
				if (!-e "$banknameout.contig_$numcontig.fasta" || -z "$banknameout.contig_$numcontig.fasta") {
					&errorMessage(__LINE__, "Cannot dump \"$primarycontig\" from \"$primarybank\" to \"$banknameout.contig_$numcontig.fasta\".");
				}
			}
			foreach my $rawmember (keys(%association)) {
				if ($association{$rawmember} eq $primarycontig) {
					$association{$rawmember} = "contig_$numcontig";
				}
			}
			$numcontig ++;
		}
	}
	# make FASTQ file for each contig
	my %nreads;
	if ($rawreads) {
		$filehandleinput1 = &readFile($rawreads);
		my $tempnline = 1;
		while (<$filehandleinput1>) {
			if ($tempnline % 4 == 1) {
				s/ nreads=\d+//;
			}
			if ($tempnline % 4 == 1 && /^\@(.+)\r?\n?$/) {
				if ($filehandleoutput1) {
					close($filehandleoutput1);
				}
				if ($singletons{$1}) {
					unless (open($filehandleoutput1, ">> $banknameout.expels.fastq")) {
						&errorMessage(__LINE__, "");
					}
				}
				elsif ($association{$1}) {
					unless (open($filehandleoutput1, ">> $banknameout.$association{$1}.fastq")) {
						&errorMessage(__LINE__, "");
					}
					$nreads{$association{$1}} ++;
				}
				else {
					&errorMessage(__LINE__, "Unknown error in processing a line\n$_\nof \"$rawreads\".");
				}
			}
			if ($filehandleoutput1) {
				if (/^.+\r?\n?$/) {
					print($filehandleoutput1 $_)
				}
			}
			else {
				&errorMessage(__LINE__, "Unknown error in processing a line\n$_\nof \"$rawreads\".");
			}
			$tempnline ++;
		}
		if ($filehandleoutput1) {
			close($filehandleoutput1);
		}
		close($filehandleinput1);
	}
	else {
		foreach my $inputfile (@firstinputfiles) {
			&dumpReadsfromBankAsFASTQ($inputfile, "$inputfile.fastq", 1);
			unless (open($filehandleinput1, "< $inputfile.fastq")) {
				&errorMessage(__LINE__, "Cannot read \"$inputfile.fastq\".");
			}
			my $tempnline = 1;
			while (<$filehandleinput1>) {
				if ($tempnline % 4 == 1 && /^\@(.+)\r?\n?$/) {
					if ($filehandleoutput1) {
						close($filehandleoutput1);
					}
					if ($singletons{$1}) {
						unless (open($filehandleoutput1, ">> $banknameout.expels.fastq")) {
							&errorMessage(__LINE__, "");
						}
					}
					elsif ($association{$1}) {
						unless (open($filehandleoutput1, ">> $banknameout.$association{$1}.fastq")) {
							&errorMessage(__LINE__, "");
						}
						$nreads{$association{$1}} ++;
					}
					else {
						&errorMessage(__LINE__, "Unknown error in processing a line\n$_\nof \"$inputfile.fastq\".");
					}
				}
				if ($filehandleoutput1) {
					if (/^.+\r?\n?$/) {
						print($filehandleoutput1 $_)
					}
				}
				else {
					&errorMessage(__LINE__, "Unknown error in processing a line\n$_\nof \"$inputfile.fastq\".");
				}
				$tempnline ++;
			}
			if ($filehandleoutput1) {
				close($filehandleoutput1);
			}
			close($filehandleinput1);
			unless (unlink("$inputfile.fastq")) {
				&errorMessage(__LINE__, "Cannot delete \"$inputfile.fastq\".");
			}
		}
	}
	my @contignames = sort({$nreads{$b} <=> $nreads{$a} || $a cmp $b} keys(%nreads));
	# execute mapping in parallel
	{
		my $child = 0;
		$| = 1;
		$? = 0;
		foreach my $contigname (@contignames) {
			if (my $pid = fork()) {
				$child ++;
				if ($child >= ($numthreads + 1)) {
					if (wait == -1) {
						$child = 0;
					} else {
						$child --;
					}
				}
				if ($?) {
					&errorMessage(__LINE__, "The processes did not finished correctly.");
				}
				next;
			}
			else {
				if ($nreads{$contigname} > 1 && -e "$banknameout.$contigname.fasta" && !-z "$banknameout.$contigname.fasta" && -e "$banknameout.$contigname.fastq" && !-z "$banknameout.$contigname.fastq") {
					&FASTQtoBank("$banknameout.$contigname.fastq", "$banknameout.$contigname.bnk", 1);
					print(STDERR "Mapping raw reads in $banknameout.$contigname.bnk...\n");
					if ($linkagemode eq 'complete') {
						&mapReadstoContig("$banknameout.$contigname", $minident3, 1);
					}
					else {
						&mapReadstoContig("$banknameout.$contigname", $minident3, 0);
					}
				}
				else {
					&errorMessage(__LINE__, "Unknown error.\ncontig : $contigname\nnreads : $nreads{$contigname}\nfastafile : $banknameout.$contigname.fasta\nfastqfile : $banknameout.$contigname.fastq");
				}
				exit;
			}
		}
		# join
		while (wait != -1) {
			if ($?) {
				&errorMessage(__LINE__, "The processes did not finished correctly.");
			}
		}
	}
	# combine banks
	{
		my @contigbank;
		foreach my $contigname (@contignames) {
			push(@contigbank, "$banknameout.$contigname.bnk");
		}
		if (scalar(@contigbank) > 1) {
			&combineBanks(@contigbank, $banknameout);
		}
		elsif (scalar(@contigbank) == 1) {
			unless (rename($contigbank[0], $banknameout)) {
				&errorMessage(__LINE__, "Cannot rename \"$contigbank[0]\" to \"$banknameout\".");
			}
		}
	}
	# check invalid files
	{
		my @tempfiles = glob("$banknameout.contig_*.fast?");
		if (@tempfiles) {
			&errorMessage(__LINE__, "Invalid files are found.\n" . join("\n", @tempfiles));
		}
	}
	# assemble expelled and singleton sequences
	if (-e $banknameout) {
		my ($contigmembers, $singletons) = &getContigMembersfromBank($banknameout);
		if ($singletons->{$banknameout} && $contigmembers->{$banknameout}) {
			print(STDERR scalar(@{$singletons->{$banknameout}}) . " sequences were expelled from contigs.\n");
			&dumpReadsfromBankAsFASTQ($banknameout, "$banknameout.expels.fastq", @{$singletons->{$banknameout}}, 1);
			&deleteReadsfromBank($banknameout, &convertReadEIDstoIIDs($banknameout, @{$singletons->{$banknameout}}));
			if (countFASTQ("$banknameout.expels.fastq") > 1) {
				print(STDERR "Reassembling expelled sequences...\n");
				my $tempbank = &makeTempFile();
				if ($linkagemode eq 'complete') {
					&reassembleFASTQ("$banknameout.expels.fastq", "$tempbank.bnk", $minident3, 0, 0, 'complete', 'nucmer', $firstaddniter);
				}
				else {
					&reassembleFASTQ("$banknameout.expels.fastq", "$tempbank.bnk", $minident3, 0, 0, 'single', 'nucmer', $firstaddniter);
				}
				if (-e "$tempbank.bnk/CTG.ifo") {
					if (-e $banknameout) {
						unless (rename($banknameout, "$banknameout.temp.bnk")) {
							&errorMessage(__LINE__, "Cannot rename \"$banknameout\" to \"$banknameout.temp.bnk\".");
						}
					}
					&deleteContigEIDs("$tempbank.bnk");
					&deleteContigEIDs("$banknameout.temp.bnk");
					&combineBanks("$banknameout.temp.bnk", "$tempbank.bnk", $banknameout);
					if (!$nodel) {
						&deleteContigMembers("$tempbank.bnk");
					}
					unless (unlink("$banknameout.expels.fastq")) {
						&errorMessage(__LINE__, "Cannot delete\"$banknameout.expels.fastq\".");
					}
				}
				else {
					&FASTQtoBank("$banknameout.expels.fastq", $banknameout, 1);
					if (!$nodel) {
						&deleteBanks("$tempbank.bnk");
						&deleteContigMembers("$tempbank.bnk");
					}
				}
			}
			else {
				&FASTQtoBank("$banknameout.expels.fastq", $banknameout, 1);
			}
		}
		elsif (-e "$banknameout.expels.fastq" && !-z "$banknameout.expels.fastq") {
			&FASTQtoBank("$banknameout.expels.fastq", $banknameout, 1);
		}
	}
	elsif (-e "$banknameout.expels.fastq" && !-z "$banknameout.expels.fastq") {
		&FASTQtoBank("$banknameout.expels.fastq", $banknameout, 1);
	}
	# delete singleton contigs
	if (-e $banknameout) {
		&deleteSingletonContigs($banknameout);
	}
	if (!$nodel) {
		&deleteBanks($banknamein);
	}
}

sub countFASTQ {
	my $filenamein = shift(@_);
	$filehandleinput1  = &readFile($filenamein);
	my $tempnline = 0;
	while (<$filehandleinput1>) {
		$tempnline ++;
	}
	close($filehandleinput1);
	return(int($tempnline / 4));
}

sub associateReadstoContig {
	my $banknamein = shift(@_);
	my $banknameout = shift(@_);
	my $contigmembers = $banknameout;
	$contigmembers =~ s/\.bnk$/.contigmembers.gz/;
	my @primarybank = @_;
	my ($secondarycontigmembers, $secondarysingletons) = &getContigMembersfromBank($banknamein);
	my ($primarycontigmembers, $primarysingletons) = &getContigMembersfromBank(@primarybank);
	my %convert;
	my $ncontig = 1;
	$filehandleoutput1 = &writeFile($contigmembers);
	if ($secondarycontigmembers->{$banknamein}) {
		foreach my $contigname (sort({$a cmp $b} keys(%{$secondarycontigmembers->{$banknamein}}))) {
			print($filehandleoutput1 $contigname);
			foreach my $member (@{$secondarycontigmembers->{$banknamein}->{$contigname}}) {
				if ($member =~ /^contig_(.+)_(\d+)$/ || ($replaceprecontig && $member =~ /^precontig_(.+)_(\d+)$/)) {
					foreach my $rawmember (@{$primarycontigmembers->{$1}->{"contig_$2"}}) {
						print($filehandleoutput1 "\t$rawmember");
					}
				}
				else {
					print($filehandleoutput1 "\t$member");
				}
			}
			print($filehandleoutput1 "\n");
			$ncontig ++;
		}
	}
	if ($secondarysingletons->{$banknamein}) {
		foreach my $member (@{$secondarysingletons->{$banknamein}}) {
			if ($member =~ /^contig_(.+)_(\d+)$/ || ($replaceprecontig && $member =~ /^precontig_(.+)_(\d+)$/)) {
				$convert{$member} = "contig_$ncontig";
				print($filehandleoutput1 $convert{$member});
				$ncontig ++;
				foreach my $rawmember (@{$primarycontigmembers->{$1}->{"contig_$2"}}) {
					print($filehandleoutput1 "\t$rawmember");
				}
				print($filehandleoutput1 "\n");
			}
			else {
				print($filehandleoutput1 "\t$member\n");
			}
		}
	}
	close($filehandleoutput1);
	&BanktoBank($banknamein, $banknameout, \%convert, 1);
}

sub makeContigMembers {
	my $bankname = shift(@_);
	my $filename = $bankname;
	$filename =~ s/\.bnk$/.contigmembers.gz/;
	if (!-e $filename) {
		my ($contigmembers, $singletons) = &getContigMembersfromBank($bankname);
		$filehandleoutput1 = &writeFile($filename);
		if ($contigmembers->{$bankname}) {
			foreach my $contigname (sort({$a cmp $b} keys(%{$contigmembers->{$bankname}}))) {
				print($filehandleoutput1 $contigname);
				foreach my $member (@{$contigmembers->{$bankname}->{$contigname}}) {
					print($filehandleoutput1 "\t$member");
				}
				print($filehandleoutput1 "\n");
			}
		}
		if ($singletons->{$bankname}) {
			foreach my $member (@{$singletons->{$bankname}}) {
				print($filehandleoutput1 "\t$member\n");
			}
		}
		close($filehandleoutput1);
	}
}

sub readFile {
	my $filehandle;
	my $filename = shift(@_);
	if ($filename =~ /\.gz$/i) {
		unless (open($filehandle, "gzip -dc $filename 2> $devnull |")) {
			&errorMessage(__LINE__, "Cannot open \"$filename\".");
		}
	}
	elsif ($filename =~ /\.bz2$/i) {
		unless (open($filehandle, "bzip2 -dc $filename 2> $devnull |")) {
			&errorMessage(__LINE__, "Cannot open \"$filename\".");
		}
	}
	elsif ($filename =~ /\.xz$/i) {
		unless (open($filehandle, "xz -dc $filename 2> $devnull |")) {
			&errorMessage(__LINE__, "Cannot open \"$filename\".");
		}
	}
	else {
		unless (open($filehandle, "< $filename")) {
			&errorMessage(__LINE__, "Cannot open \"$filename\".");
		}
	}
	return($filehandle);
}

sub writeFile {
	my $filehandle;
	my $filename = shift(@_);
	if ($filename =~ /\.gz$/i) {
		unless (open($filehandle, "| gzip -c >> $filename 2> $devnull")) {
			&errorMessage(__LINE__, "Cannot open \"$filename\".");
		}
	}
	elsif ($filename =~ /\.bz2$/i) {
		unless (open($filehandle, "| bzip2 -c >> $filename 2> $devnull")) {
			&errorMessage(__LINE__, "Cannot open \"$filename\".");
		}
	}
	elsif ($filename =~ /\.xz$/i) {
		unless (open($filehandle, "| xz -c >> $filename 2> $devnull")) {
			&errorMessage(__LINE__, "Cannot open \"$filename\".");
		}
	}
	else {
		unless (open($filehandle, "> $filename")) {
			&errorMessage(__LINE__, "Cannot open \"$filename\".");
		}
	}
	return($filehandle);
}

sub BanktoBank {
	my $banknamein = shift(@_);
	my $banknameout = shift(@_);
	my $convert = shift(@_);
	my $deletefile = shift(@_);
	my $pipehandleinput1;
	my $pipehandleoutput1;
	unless (open($pipehandleinput1, "$bankreport -p -b $banknamein 2> $devnull |")) {
		&errorMessage(__LINE__, "Cannot run \"$bankreport -p -b $banknamein\".");
	}
	unless (open($pipehandleoutput1, "| $banktransact -c -b $banknameout -m - 1> $devnull 2> $devnull")) {
		&errorMessage(__LINE__, "Cannot run \"$banktransact -c -b $banknameout -m -\".");
	}
	my $temp;
	my $nesting = 0;
	my $read;
	my $contig;
	while (<$pipehandleinput1>) {
		if ($nesting && /^\}/) {
			$nesting --;
		}
		elsif (/^\{([A-Z]+)/) {
			$nesting ++;
			if ($1 eq 'CTG') {
				$contig ++;
			}
			elsif ($1 eq 'RED') {
				$read ++;
			}
		}
		$temp .= $_;
		if ($nesting == 0) {
			if ($contig) {
				$temp =~ s/^eid:\S+\r?\n?//m;
				unless ($noeid) {
					$temp =~ s/^iid:(\d+)/$&\neid:contig_$1/m;
				}
			}
			if ($read && $temp =~ /^eid:(contig_.+_\d+)\r?\n?$/m && $convert->{$1}) {
				$temp =~ s/^eid:(contig_.+_\d+)(\r?\n?)$/eid:$convert->{$1}$2/m;
			}
			elsif ($replaceprecontig && $read && $temp =~ /^eid:(precontig_.+_\d+)\r?\n?$/m && $convert->{$1}) {
				$temp =~ s/^eid:(precontig_.+_\d+)(\r?\n?)$/eid:$convert->{$1}$2/m;
			}
			print($pipehandleoutput1 $temp);
			undef($temp);
			undef($read);
			undef($contig);
		}
	}
	close($pipehandleoutput1);
	close($pipehandleinput1);
	if (!-e "$banknameout/CTG.ifo" && -e "$banknamein/CTG.ifo") {
		my @ctg = glob("$banknamein/CTG.*");
		foreach my $ctgfile (@ctg) {
			$ctgfile =~ /[\\\/]([^\\\/]+)$/;
			my $tempfilename = $1;
			unless (fcopy($ctgfile, "$banknameout/$tempfilename")) {
				&errorMessage(__LINE__, "Cannot move \"$ctgfile\" to \"$banknameout/$tempfilename\".");
			}
		}
	}
	if ($deletefile) {
		if (!$nodel) {
			&deleteBanks($banknamein);
		}
	}
}

sub FASTQtoBank {
	my $filename = shift(@_);
	my $bankname = shift(@_);
	my $deletefile = shift(@_);
	if (system("$toAmos_new -Q $filename -t SANGER -b $bankname 2> $devnull 1> $devnull")) {
		&errorMessage(__LINE__, "Cannot run \"$toAmos_new -Q $filename -t SANGER -b $bankname\".");
	}
	if ($deletefile) {
		if (!$nodel) {
			unless (unlink($filename)) {
				&errorMessage(__LINE__, "Cannot delete \"$filename\".");
			}
		}
	}
}

sub mapReadstoContig {
	my $prefix = shift(@_);
	my $minident = shift(@_);
	my $expel = shift(@_);
	my $maxconerr;
	my $tempminovllen;
	if ($minovllen == 0) {
		$tempminovllen = &getMinimumLength("$prefix.bnk") - 10;
		if ($tempminovllen < 50) {
			$tempminovllen = 50;
		}
	}
	else {
		$tempminovllen = $minovllen;
	}
	my $prec = length($tempminovllen) + 1;
	if ($minident == 0) {
		$minident = &truncateValue(($tempminovllen - 1) / $tempminovllen, $prec);
	}
	if ($expel) {
		$maxconerr = (1 - $minident) * 2;
	}
	else {
		$maxconerr = 0.6;
	}
	# map reads to contig
	if (-e "$prefix.fasta" && !-e "$prefix.bnk/LAY.ifo" && !-e "$prefix.bnk/CTG.ifo") {
		&dumpReadsfromBankAsFASTA("$prefix.bnk", "$prefix.temp", 0);
		if ($strand) {
			if (system("$nucmer --forward --maxmatch --delta --extend --optimize --simplify --breaklen 200 --mincluster 65 --diagfactor 0.12 --maxgap 90 --minmatch 7 --prefix=$prefix $prefix.fasta $prefix.temp 2> $devnull 1> $devnull")) {
				&errorMessage(__LINE__, "Cannot run \"$nucmer --forward --maxmatch --delta --extend --optimize --simplify --breaklen 200 --mincluster 65 --diagfactor 0.12 --maxgap 90 --minmatch 7 --prefix=$prefix $prefix.fasta $prefix.temp\".");
			}
		}
		else {
			if (system("$nucmer --maxmatch --delta --extend --optimize --simplify --breaklen 200 --mincluster 65 --diagfactor 0.12 --maxgap 90 --minmatch 7 --prefix=$prefix $prefix.fasta $prefix.temp 2> $devnull 1> $devnull")) {
				&errorMessage(__LINE__, "Cannot run \"$nucmer --maxmatch --delta --extend --optimize --simplify --breaklen 200 --mincluster 65 --diagfactor 0.12 --maxgap 90 --minmatch 7 --prefix=$prefix $prefix.fasta $prefix.temp\".");
			}
		}
		unless (unlink("$prefix.temp")) {
			&errorMessage(__LINE__, "Cannot delete \"$prefix.temp\".");
		}
		if (!$nodel) {
			unless (unlink("$prefix.fasta")) {
				&errorMessage(__LINE__, "Cannot delete \"$prefix.fasta\".");
			}
		}
	}
	# layout
	if (!-e "$prefix.bnk/LAY.ifo" && !-e "$prefix.bnk/CTG.ifo") {
		if (system("$casmlayout -t 100 -b $prefix.bnk $prefix.delta 2> $devnull 1> $devnull")) {
			&errorMessage(__LINE__, "Cannot run \"$casmlayout -b $prefix.bnk $prefix.delta\".");
		}
		if (!$nodel) {
			unless (unlink("$prefix.delta")) {
				&errorMessage(__LINE__, "Cannot delete \"$prefix.delta\".");
			}
		}
	}
	# consensus
	if (!-e "$prefix.bnk/CTG.ifo") {
		if (system("$makeconsensus -B -b -o $tempminovllen -e $maxconerr -w $alnwiggle -v 0 $prefix.bnk 2> $devnull 1> $devnull")) {
			&errorMessage(__LINE__, "Cannot run \"$makeconsensus -B -b -o $tempminovllen -e $maxconerr -w $alnwiggle -v 0 $prefix.bnk\".");
		}
	}
	# remake consensus
	if (-e "$prefix.bnk/CTG.ifo") {
		if (system("$recallConsensus -b $prefix.bnk 2> $devnull 1> $devnull")) {
			system("$bankunlock $prefix.bnk 2> $devnull 1> $devnull");
		}
	}
	else {
		&errorMessage(__LINE__, "\"$prefix.bnk/CTG.ifo\" does not exist.");
	}
	&deleteContigEIDs("$prefix.bnk");
}

sub assembleBank {
	my $bankname = shift(@_);
	my $minident = shift(@_);
	my $expel = shift(@_);
	my $mapping = shift(@_);
	my $nseqfirst = shift(@_);
	my $maxovlerr;
	my $maxconerr;
	my $tempminovllen;
	if ($minovllen == 0) {
		$tempminovllen = &getMinimumLength($bankname) - 10;
		if ($tempminovllen < 50) {
			$tempminovllen = 50;
		}
	}
	else {
		$tempminovllen = $minovllen;
	}
	my $prec = length($tempminovllen) + 1;
	if ($minident == 0) {
		$minident = &truncateValue(($tempminovllen - 1) / $tempminovllen, $prec);
	}
	if ($mapping) {
		$maxovlerr = 0.5;
		$maxconerr = (1 - $minident) * 2;
	}
	elsif ($expel) {
		$maxovlerr = 1 - $minident;
		$maxconerr = (1 - $minident) * 2;
	}
	else {
		$maxovlerr = 1 - $minident;
		$maxconerr = 0.6;
	}
	my $overlapper;
	if ($mapping) {
		$overlapper = $mappingmode;
		$nseqfirst = 1;
	}
	else {
		$overlapper = $assemblemode;
	}
	# detect overlap
	if (!-e "$bankname/OVL.ifo" && !-e "$bankname/LAY.ifo" && !-e "$bankname/CTG.ifo") {
		if ($overlapper eq 'blat' || $overlapper eq 'blatfast') {
			if ($nseqfirst) {
				my $tempnseq = 0;
				unless (open($pipehandleinput1, "$dumpreads -r $bankname 2> $devnull |")) {
					&errorMessage(__LINE__, "Cannot run \"$dumpreads -r $bankname\".");
				}
				unless (open($filehandleoutput1, "> $bankname.temp1.fasta")) {
					&errorMessage(__LINE__, "Cannot make \"$bankname.temp1.fasta\".");
				}
				while (<$pipehandleinput1>) {
					s/\r?\n?$//;
					if (/^>/) {
						$tempnseq ++;
						if ($tempnseq == $nseqfirst + 1) {
							print($filehandleoutput1 "\n");
							close($filehandleoutput1);
							unless (open($filehandleoutput1, "> $bankname.temp2.fasta")) {
								&errorMessage(__LINE__, "Cannot make \"$bankname.temp2.fasta\".");
							}
							print($filehandleoutput1 $_ . "\n");
						}
						elsif ($tempnseq == 1) {
							print($filehandleoutput1 $_ . "\n");
						}
						else {
							print($filehandleoutput1 "\n" . $_ . "\n");
						}
					}
					else {
						print($filehandleoutput1 $_);
					}
				}
				print($filehandleoutput1 "\n");
				close($filehandleoutput1);
				close($pipehandleinput1);
				if ($?) {
					&errorMessage(__LINE__, "Cannot run \"$dumpreads -r $bankname\".");
				}
			}
			elsif (system("$dumpreads -r $bankname 2> $devnull > $bankname.temp1.fasta")) {
				&errorMessage(__LINE__, "Cannot run \"$dumpreads -r $bankname > $bankname.temp1.fasta\".");
			}
			my $pblatoption = " -threads=$numthreads -noHead -noTrimA -out=psl -maxIntron=0 -repMatch=2147483647 -t=dna -q=dna";
			if ($overlapper eq 'blatfast') {
				$pblatoption .= ' -fastMap';
			}
			if ($strand) {
				$pblatoption .= ' -strand=plus';
			}
			else {
				$pblatoption .= ' -strand=both';
			}
			my $tilesize;
			my $maxgap;
			my $oneoff;
			my $minscore;
			if ($minident == 1) {
				$maxgap = 0;
				$oneoff = 0;
				$tilesize = 18;
				$minscore = $tempminovllen;
			}
			else {
				if ($minident >= 0.97) {
					$maxgap = 1;
				}
				elsif ($minident >= 0.93) {
					$maxgap = 2;
				}
				else {
					$maxgap = 3;
				}
				$oneoff = 1;
				$tilesize = int($minident * 100) - 85;
				if ($tilesize < 8) {
					$tilesize = 8;
				}
				$minscore = int(($tempminovllen * $minident) - ($tempminovllen - ($tempminovllen * $minident)) * 2);
				if ($minscore < $tempminovllen / 2) {
					$minscore = int($tempminovllen / 2);
				}
			}
			$pblatoption .= " -minScore=$minscore -oneOff=$oneoff -minIdentity=" . sprintf("%.*f", $prec - 2, $minident * 100) . " -maxGap=$maxgap -tileSize=$tilesize -stepSize=$tilesize -minMatch=" . int($tempminovllen / $tilesize);
			#print(STDERR "The minimum overlap length is shifted to " . ($tilesize * int($tempminovllen / $tilesize)) . " because of algorithm limitation.\n");
			if (-e "$bankname.temp2.fasta") {
				if (system("$pblat$pblatoption $bankname.temp1.fasta $bankname.temp2.fasta $bankname.psl 1> $devnull 2> $devnull")) {
					&errorMessage(__LINE__, "Cannot run \"$pblat$pblatoption $bankname.temp1.fasta $bankname.temp2.fasta $bankname.psl\".");
				}
			}
			else {
				if (system("$pblat$pblatoption $bankname.temp1.fasta $bankname.temp1.fasta $bankname.psl 1> $devnull 2> $devnull")) {
					&errorMessage(__LINE__, "Cannot run \"$pblat$pblatoption $bankname.temp1.fasta $bankname.temp1.fasta $bankname.psl\".");
				}
			}
			if (system("$blat2nucmer $bankname.psl 2> $devnull | $nucmerAnnotate -ignore 0 -noid -loose 2> $devnull | grep -P '\[(?:BEGIN|END|CONTAINED|CONTAINS|IDENTITY)\]' | $nucmer2ovl -ignore 2 -tab 2> $devnull > $bankname.ovl")) {
				&errorMessage(__LINE__, "Cannot run \"$blat2nucmer $bankname.psl | $nucmerAnnotate -ignore 0 -noid -loose | grep -P '\[(?:BEGIN|END|CONTAINED|CONTAINS|IDENTITY)\]' | $nucmer2ovl -ignore 2 -tab > $bankname.ovl\".");
			}
			if (-e "$bankname.temp1.fasta") {
				unless (unlink("$bankname.temp1.fasta")) {
					&errorMessage(__LINE__, "Cannot delete \"$bankname.temp1.fasta\".");
				}
			}
			if (-e "$bankname.temp2.fasta") {
				if (!$nodel) {
					unless (unlink("$bankname.temp2.fasta")) {
						&errorMessage(__LINE__, "Cannot delete \"$bankname.temp2.fasta\".");
					}
				}
			}
			if (-e "$bankname.psl") {
				if (!$nodel) {
					unless (unlink("$bankname.psl")) {
						&errorMessage(__LINE__, "Cannot delete \"$bankname.psl\".");
					}
				}
			}
		}
		elsif ($overlapper eq 'minimus') {
			if ($nseqfirst) {
				$nseqfirst = " -r $nseqfirst";
			}
			else {
				$nseqfirst = '';
			}
			if ($mapping) {
				if (system("OMP_NUM_THREADS=$numthreads $hashoverlap$nseqfirst$strand -o $tempminovllen -x $maxovlerr -v 0 -B $bankname > $devnull 2> $devnull")) {
					&errorMessage(__LINE__, "Cannot run \"OMP_NUM_THREADS=$numthreads $hashoverlap$nseqfirst$strand -o $tempminovllen -x $maxovlerr -v 0 -B $bankname\".");
				}
			}
			else {
				if (system("OMP_NUM_THREADS=$numthreads $hashoverlap$nseqfirst$strand -o $tempminovllen -x $maxovlerr -v 0 $bankname > $bankname.ovl 2> $devnull")) {
					&errorMessage(__LINE__, "Cannot run \"OMP_NUM_THREADS=$numthreads $hashoverlap$nseqfirst$strand -o $tempminovllen -x $maxovlerr -v 0 $bankname > $bankname.ovl\".");
				}
			}
		}
		# sort contents of ovl file
		if (!-e "$bankname/OVL.ifo") {
			if (-e "$bankname.ovl" && !-z "$bankname.ovl") {
				if (system("LANG=C LC_ALL=C sort -b -k 1,2 -t \"	\" -n $bankname.ovl 2> $devnull | $ovl2OVL 2> $devnull | $banktransact -b $bankname -m - 2> $devnull 1> $devnull")) {
					&errorMessage(__LINE__, "Cannot run \"LANG=C LC_ALL=C sort -b -k 1,2 -t \"	\" -n $bankname.ovl | $ovl2OVL | $banktransact -b $bankname -m -\".");
				}
				if (!$nodel) {
					unless (unlink("$bankname.ovl")) {
						&errorMessage(__LINE__, "Cannot delete \"$bankname.ovl\".");
					}
				}
			}
			elsif (-e "$bankname.ovl") {
				unless (open($pipehandleoutput1, "| $banktransact -b $bankname -m - 2> $devnull 1> $devnull")) {
					&errorMessage(__LINE__, "Cannot run \"$banktransact -b $bankname -m -\".");
				}
				print($pipehandleoutput1 "{OVL\nadj:N\nscr:0\nrds:0,0\nahg:0\nbhg:0\n}\n");
				close($pipehandleoutput1);
				if ($?) {
					&errorMessage(__LINE__, "Cannot run \"$banktransact -b $bankname -m -\".");
				}
				if (!$nodel) {
					unless (unlink("$bankname.ovl")) {
						&errorMessage(__LINE__, "Cannot delete \"$bankname.ovl\".");
					}
				}
			}
			else {
				&errorMessage(__LINE__, "Cannot find \"$bankname.ovl\".");
			}
		}
	}
	# layout
	if (-e "$bankname/OVL.ifo" && !-e "$bankname/LAY.ifo" && !-e "$bankname/CTG.ifo") {
		if (system("$tigger -v 0 -b $bankname 2> $devnull 1> $devnull")) {
			&errorMessage(__LINE__, "Cannot run \"$tigger -v 0 -b $bankname\".");
		}
	}
	# consensus
	if (-e "$bankname/LAY.ifo" && !-e "$bankname/CTG.ifo") {
		if (system("$makeconsensus -b -B -o $tempminovllen -e $maxconerr -w $alnwiggle -v 0 -x $bankname.expels.iid $bankname 2> $devnull 1> $devnull")) {
			&errorMessage(__LINE__, "Cannot run \"$makeconsensus -b -B -o $tempminovllen -e $maxconerr -w $alnwiggle -v 0 -x $bankname.expels.iid $bankname\".");
		}
	}
	if ($mapping) {
		if (!-e "$bankname/CTG.ifo") {
			&errorMessage(__LINE__, "\"$bankname/CTG.ifo\" does not exist.");
		}
		&cleanBank($bankname, 'CTG', 'RED');
		if (!-e "$bankname/CTG.ifo") {
			&errorMessage(__LINE__, "\"$bankname/CTG.ifo\" does not exist.");
		}
		&deleteReadsfromBank($bankname, 1);
		if (!-e "$bankname/CTG.ifo") {
			&errorMessage(__LINE__, "\"$bankname/CTG.ifo\" does not exist.");
		}
	}
	# make expelled sequence file ## need to rewrite (add singletons in mapping)
	if ($expel == 2 && -e "$bankname.expels.iid" && !-z "$bankname.expels.iid") {
		# read expel file
		my @expels;
		unless (open($filehandleinput1, "< $bankname.expels.iid")) {
			&errorMessage(__LINE__, "Cannot read \"$bankname.expels.iid\".");
		}
		while (<$filehandleinput1>) {
			if (/^(\d+)\s+\d+/ && $1 > 1) {
				push(@expels, $1);
			}
		}
		close($filehandleinput1);
		if (@expels) {
			&dumpReadsfromBankAsFASTQ($bankname, "$outputbank.expels.fastq", @expels, 1);
			&deleteReadsfromBank($bankname, @expels);
		}
	}
	if (-e "$bankname.expels.iid") {
		unless (unlink("$bankname.expels.iid")) {
			&errorMessage(__LINE__, "Cannot delete \"$bankname.expels.iid\"");
		}
	}
	# remake consensus
	if (-e "$bankname/CTG.ifo") {
		if (system("$recallConsensus -b $bankname 2> $devnull 1> $devnull")) {
			system("$bankunlock $bankname 2> $devnull 1> $devnull");
		}
	}
	else {
		&errorMessage(__LINE__, "\"$bankname/CTG.ifo\" does not exist.");
	}
}

sub makeTempFile {
	my $tempnum = 1;
	my $tempfilename = sprintf("%06d", $tempnum);
	my @tempfiles = glob("$tempfilename.*");
	while (-e $tempfilename || scalar(@tempfiles) > 0) {
		$tempnum ++;
		$tempfilename = sprintf("%06d", $tempnum);
		@tempfiles = glob("$tempfilename.*");
	}
	return($tempfilename);
}

sub deleteContigEIDs {
	my $bankname = shift(@_);
	&cleanBank($bankname, 'CTG', 'RED');
	unless (open($pipehandleinput1, "$bankreport -p -b $bankname 2> $devnull |")) {
		&errorMessage(__LINE__, "Cannot run \"$bankreport -p -b $bankname\".");
	}
	unless (open($pipehandleoutput1, "| $banktransact -c -f -b $bankname.temporary -m - 1> $devnull 2> $devnull")) {
		&errorMessage(__LINE__, "Cannot run \"$banktransact -c -f -b $bankname.temporary -m -\".");
	}
	my $temp;
	my $nesting = 0;
	my $contig;
	while (<$pipehandleinput1>) {
		if ($nesting && /^\}/) {
			$nesting --;
		}
		elsif (/^\{([A-Z]+)/) {
			$nesting ++;
			if ($1 eq 'CTG') {
				$contig ++;
			}
		}
		$temp .= $_;
		if ($nesting == 0) {
			if ($contig) {
				$temp =~ s/^eid:\S+\r?\n?//m;
			}
			print($pipehandleoutput1 $temp);
			undef($temp);
			undef($contig);
		}
	}
	close($pipehandleoutput1);
	if ($?) {
		&errorMessage(__LINE__, "Cannot run \"$banktransact -c -f -b $bankname.temporary -m -\".");
	}
	close($pipehandleinput1);
	if ($?) {
		&errorMessage(__LINE__, "Cannot run \"$bankreport -p -b $bankname\".");
	}
	if (!-e "$bankname.temporary/CTG.ifo" && -e "$bankname/CTG.ifo") {
		my @ctg = glob("$bankname/CTG.*");
		foreach my $ctgfile (@ctg) {
			$ctgfile =~ /[\\\/]([^\\\/]+)$/;
			my $tempfilename = $1;
			unless (fcopy($ctgfile, "$bankname.temporary/$tempfilename")) {
				&errorMessage(__LINE__, "Cannot copy \"$ctgfile\" to \"$bankname.temporary/$tempfilename\".");
			}
		}
	}
	# delete old bank
	&deleteBanks($bankname);
	# rename new bank
	unless (rename("$bankname.temporary", $bankname)) {
		&errorMessage(__LINE__, "Cannot rename \"$bankname.temporary\" to \"$bankname\".");
	}
}

sub addEIDstoContigs {
	my $bankname = shift(@_);
	&cleanBank($bankname, 'CTG', 'RED');
	unless (open($pipehandleinput1, "$bankreport -p -b $bankname 2> $devnull |")) {
		&errorMessage(__LINE__, "Cannot run \"$bankreport -p -b $bankname\".");
	}
	unless (open($pipehandleoutput1, "| $banktransact -c -f -b $bankname.temporary -m - 1> $devnull 2> $devnull")) {
		&errorMessage(__LINE__, "Cannot run \"$banktransact -c -f -b $bankname.temporary -m -\".");
	}
	my $temp;
	my $nesting = 0;
	my $contig;
	while (<$pipehandleinput1>) {
		if ($nesting && /^\}/) {
			$nesting --;
		}
		elsif (/^\{([A-Z]+)/) {
			$nesting ++;
			if ($1 eq 'CTG') {
				$contig ++;
			}
		}
		$temp .= $_;
		if ($nesting == 0) {
			if ($contig) {
				$temp =~ s/^eid:\S+\r?\n?//m;
				unless ($noeid) {
					$temp =~ s/^iid:(\d+)/$&\neid:contig_$1/m;
				}
			}
			print($pipehandleoutput1 $temp);
			undef($temp);
			undef($contig);
		}
	}
	close($pipehandleoutput1);
	if ($?) {
		&errorMessage(__LINE__, "Cannot run \"$banktransact -c -f -b $bankname.temporary -m -\".");
	}
	close($pipehandleinput1);
	if ($?) {
		&errorMessage(__LINE__, "Cannot run \"$bankreport -p -b $bankname\".");
	}
	if (!-e "$bankname.temporary/CTG.ifo" && -e "$bankname/CTG.ifo") {
		my @ctg = glob("$bankname/CTG.*");
		foreach my $ctgfile (@ctg) {
			$ctgfile =~ /[\\\/]([^\\\/]+)$/;
			my $tempfilename = $1;
			unless (fcopy($ctgfile, "$bankname.temporary/$tempfilename")) {
				&errorMessage(__LINE__, "Cannot copy \"$ctgfile\" to \"$bankname.temporary/$tempfilename\".");
			}
		}
	}
	# delete old bank
	&deleteBanks($bankname);
	# rename new bank
	unless (rename("$bankname.temporary", $bankname)) {
		&errorMessage(__LINE__, "Cannot rename \"$bankname.temporary\" to \"$bankname\".");
	}
}

sub truncateValue {
    my $value = shift(@_);
    my $digit = shift(@_);
    my $tempval = 10 ** $digit;
    return(int($value * $tempval) / $tempval);
}

sub reassembleFASTQ {
	my $filename = shift(@_);
	my $bankname = shift(@_);
	my $minident = shift(@_);
	my $deletefile = shift(@_);
	my $tempdepth = shift(@_);
	my $linkage = shift(@_);
	my $mapping = shift(@_);
	my $tempaddniter = shift(@_);
	# re-assemble
	my $tempopt;
	if ($strand) {
		$tempopt .= " --strand=plus";
	}
	if ($maxnsplit) {
		$tempopt .= " --maxnsplits=$maxnsplit";
	}
	if ($nodel) {
		$tempopt .= " --nodel";
	}
	if ($tempaddniter) {
		$tempopt .= " --addniter=$tempaddniter";
	}
	if (system("$assams$tempopt --assemblemode=$assemblemode --mappingmode=$mapping --linkagemode=$linkage --mergemode=$mergemode --procdepth=$tempdepth --noeid --minnseq=$minnseq --minident=$minident --minovllen=$minovllen --alnwiggle=$alnwiggle --numthreads=$numthreads $filename $bankname")) {
		&errorMessage(__LINE__, "Cannot run \"$assams$tempopt --assemblemode=$assemblemode --mappingmode=$mapping --linkagemode=$linkage --mergemode=$mergemode --procdepth=$tempdepth --noeid --minnseq=$minnseq --minident=$minident --minovllen=$minovllen --alnwiggle=$alnwiggle --numthreads=$numthreads $filename $bankname\".");
	}
	if ($deletefile) {
		if (!$nodel) {
			unless (unlink($filename)) {
				&errorMessage(__LINE__, "Cannot delete \"$filename\"");
			}
		}
	}
}

sub getMinimumLength {
	my $bankname = shift(@_);
	my $minlen;
	unless (open($pipehandleinput1, "$dumpreads -r $bankname 2> $devnull |")) {
		&errorMessage(__LINE__, "Cannot run \"$dumpreads -r $bankname\".");
	}
	my $templength = 0;
	while (<$pipehandleinput1>) {
		s/\r?\n?$//;
		s/\s+//g;
		if (/^>/ && $templength > 0) {
			if (!defined($minlen) || $templength < $minlen) {
				$minlen = $templength;
			}
			$templength = 0;
		}
		else {
			$templength += length($_);
		}
	}
	close($pipehandleinput1);
	if (!defined($minlen) || $templength < $minlen) {
		$minlen = $templength;
	}
	if ($minlen < 55) {
		$minlen = 55;
	}
	if ($minlen > 10000) {
		$minlen = 10000;
	}
	return($minlen);
}

sub dumpReadsfromBankAsFASTQ {
	my $bankname = shift(@_);
	my $filename = shift(@_);
	my $labelsareeids = pop(@_);
	my @temp = @_;
	my $option = ' -I';
	if (@temp) {
		unless (open($filehandleoutput1, "> $bankname.temp.iid")) {
			&errorMessage(__LINE__, "Cannot make \"$bankname.temp.iid\".");
		}
		foreach my $iid (@temp) {
			if ($iid !~ /^\d+$/) {
				$option = ' -E';
			}
			print($filehandleoutput1 "$iid\n");
		}
		close($filehandleoutput1);
		$option .= " $bankname.temp.iid"
	}
	else {
		$option = '';
	}
	if ($labelsareeids) {
		$option = '-e' . $option;
	}
	if (system("$dumpreads -f -Q 33 -r $option $bankname >> $filename 2> $devnull")) {
		&errorMessage(__LINE__, "Cannot run \"$dumpreads -f -Q 33 -r $option $bankname >> $filename\".");
	}
	if (-e "$bankname.temp.iid") {
		unless (unlink("$bankname.temp.iid")) {
			&errorMessage(__LINE__, "Cannot delete \"$bankname.temp.iid\".");
		}
	}
}

sub dumpReadsfromBankAsFASTA {
	my $bankname = shift(@_);
	my $filename = shift(@_);
	my $labelsareeids = pop(@_);
	my @temp = @_;
	my $option = ' -I';
	if (@temp) {
		unless (open($filehandleoutput1, "> $bankname.temp.iid")) {
			&errorMessage(__LINE__, "Cannot make \"$bankname.temp.iid\".");
		}
		foreach my $iid (@temp) {
			if ($iid !~ /^\d+$/) {
				$option = ' -E';
			}
			print($filehandleoutput1 "$iid\n");
		}
		close($filehandleoutput1);
		$option .= " $bankname.temp.iid"
	}
	else {
		$option = '';
	}
	if ($labelsareeids) {
		$option = '-e' . $option;
	}
	if (system("$dumpreads -r $option $bankname >> $filename 2> $devnull")) {
		&errorMessage(__LINE__, "Cannot run \"$dumpreads -r $option $bankname >> $filename\".");
	}
	if (-e "$bankname.temp.iid") {
		unless (unlink("$bankname.temp.iid")) {
			&errorMessage(__LINE__, "Cannot delete \"$bankname.temp.iid\".");
		}
	}
}

sub dumpContigsfromBankAsFASTQ {
	my $bankname = shift(@_);
	my $filename = shift(@_);
	my @temp = @_;
	my $option = ' -I';
	if (@temp) {
		unless (open($filehandleoutput1, "> $bankname.temp.iid")) {
			&errorMessage(__LINE__, "Cannot make \"$bankname.temp.iid\".");
		}
		foreach my $iid (@temp) {
			if ($iid !~ /^\d+$/) {
				$option = ' -E';
			}
			print($filehandleoutput1 "$iid\n");
		}
		close($filehandleoutput1);
		$option .= " $bankname.temp.iid"
	}
	else {
		$option = '';
	}
	unless (open($pipehandleinput1, "$bank2fastq -Q 33 -iid$option -b $bankname 2> $devnull |")) {
		&errorMessage(__LINE__, "Cannot run \"$bank2fastq -Q 33 -iid$option -b $bankname\".");
	}
	unless (open($filehandleoutput1, ">> $filename")) {
		&errorMessage(__LINE__, "Cannot write \"$filename\".");
	}
	my $tempnline = 1;
	while (<$pipehandleinput1>) {
		if ($tempnline % 4 == 1 && /^\@/) {
			s/^\@(\d+)(\r?\n?)$/\@contig_$1$2/;
		}
		print($filehandleoutput1 $_);
		$tempnline ++;
	}
	close($filehandleoutput1);
	close($pipehandleinput1);
	#if (($option && -z $filename) || $?) {
	#	&errorMessage(__LINE__, "Cannot run \"$bank2fastq -Q 33 -iid$option -b $bankname\".");
	#}
	if (-e "$bankname.temp.iid") {
		unless (unlink("$bankname.temp.iid")) {
			&errorMessage(__LINE__, "Cannot delete \"$bankname.temp.iid\".");
		}
	}
	#unless ($option) {
	#	unless (open($pipehandleinput1, "$dumpreads -f -Q 33 -r -e $bankname 2> $devnull |")) {
	#		&errorMessage(__LINE__, "Cannot run \"$dumpreads -f -Q 33 -r -e $bankname\".");
	#	}
	#	unless (open($filehandleoutput1, ">> $filename")) {
	#		&errorMessage(__LINE__, "Cannot write \"$filename\".");
	#	}
	#	my $tempnline = 1;
	#	my $output = 0;
	#	while (<$pipehandleinput1>) {
	#		if ($tempnline % 4 == 1 && /^\@contig_\d+\r?\n?$/) {
	#			$output = 1;
	#		}
	#		elsif ($tempnline % 4 == 1) {
	#			$output = 0;
	#		}
	#		if ($output) {
	#			print($filehandleoutput1 $_);
	#		}
	#		$tempnline ++;
	#	}
	#	close($filehandleoutput1);
	#	close($pipehandleinput1);
	#}
}

sub dumpContigsfromBankAsFASTA {
	my $bankname = shift(@_);
	my $filename = shift(@_);
	my @temp = @_;
	my $option = ' -I';
	if (@temp) {
		unless (open($filehandleoutput1, "> $bankname.temp.iid")) {
			&errorMessage(__LINE__, "Cannot make \"$bankname.temp.iid\".");
		}
		foreach my $iid (@temp) {
			if ($iid !~ /^\d+$/) {
				$option = ' -E';
			}
			print($filehandleoutput1 "$iid\n");
		}
		close($filehandleoutput1);
		$option .= " $bankname.temp.iid"
	}
	else {
		$option = '';
	}
	unless (open($pipehandleinput1, "$bank2fasta -iid$option -b $bankname 2> $devnull |")) {
		&errorMessage(__LINE__, "Cannot run \"$bank2fasta -iid$option -b $bankname\".");
	}
	unless (open($filehandleoutput1, ">> $filename")) {
		&errorMessage(__LINE__, "Cannot write \"$filename\".");
	}
	while (<$pipehandleinput1>) {
		if (/^>/) {
			s/^>(\d+)(\r?\n?)$/>contig_$1$2/;
		}
		print($filehandleoutput1 $_);
	}
	close($filehandleoutput1);
	close($pipehandleinput1);
	#if (($option && -z $filename) || $?) {
	#	&errorMessage(__LINE__, "Cannot run \"$bank2fasta -iid$option -b $bankname\".");
	#}
	if (-e "$bankname.temp.iid") {
		unless (unlink("$bankname.temp.iid")) {
			&errorMessage(__LINE__, "Cannot delete \"$bankname.temp.iid\".");
		}
	}
	#unless ($option) {
	#	unless (open($pipehandleinput1, "$dumpreads -r -e $bankname 2> $devnull |")) {
	#		&errorMessage(__LINE__, "Cannot run \"$dumpreads -r -e $bankname\".");
	#	}
	#	unless (open($filehandleoutput1, ">> $filename")) {
	#		&errorMessage(__LINE__, "Cannot write \"$filename\".");
	#	}
	#	my $output = 0;
	#	while (<$pipehandleinput1>) {
	#		if (/^>contig_\d+\r?\n?$/) {
	#			$output = 1;
	#		}
	#		elsif (/^>/) {
	#			$output = 0;
	#		}
	#		if ($output) {
	#			print($filehandleoutput1 $_);
	#		}
	#	}
	#	close($filehandleoutput1);
	#	close($pipehandleinput1);
	#}
}

sub convertReadEIDstoIIDs {
	my $bankname = shift(@_);
	my @temp = @_;
	my $option = ' -E';
	my @iids;
	if (@temp) {
		unless (open($filehandleoutput1, "> $bankname.temp.eid")) {
			&errorMessage(__LINE__, "Cannot make \"$bankname.temp.eid\".");
		}
		foreach my $eid (@temp) {
			print($filehandleoutput1 "$eid\n");
		}
		close($filehandleoutput1);
		$option .= " $bankname.temp.eid"
	}
	else {
		&errorMessage(__LINE__, "EIDs were not given.");
	}
	unless (open($pipehandleinput1, "$dumpreads -r $option $bankname 2> $devnull |")) {
		&errorMessage(__LINE__, "Cannot run \"$dumpreads -r $option $bankname\".");
	}
	while (<$pipehandleinput1>) {
		if (/^>(\d+)/) {
			push(@iids, $1);
		}
	}
	close($pipehandleinput1);
	if ($?) {
		&errorMessage(__LINE__, "Cannot run \"$dumpreads -r $option $bankname\".");
	}
	unless (unlink("$bankname.temp.eid")) {
		&errorMessage(__LINE__, "Cannot delete \"$bankname.temp.eid\".");
	}
	return(@iids);
}

sub deleteReadsfromBank {
	my $bankname = shift(@_);
	my @temp = @_;
	my %iids;
	foreach my $iid (@temp) {
		$iids{$iid} = 1;
	}
	unless (open($pipehandleinput1, "$bankreport -p -b $bankname 2> $devnull |")) {
		&errorMessage(__LINE__, "Cannot run \"$bankreport -p -b $bankname\".");
	}
	unless (open($pipehandleoutput1, "| $banktransact -c -f -b $bankname.temp.bnk -m - 1> $devnull 2> $devnull")) {
		&errorMessage(__LINE__, "Cannot run \"$banktransact -c -f -b $bankname.temp.bnk -m -\".");
	}
	my $temp;
	my $nesting = 0;
	my $tle;
	my $read;
	while (<$pipehandleinput1>) {
		if ($nesting && /^\}/) {
			$nesting --;
		}
		elsif (/^\{(RED|TLE)/) {
			$nesting ++;
			if ($1 eq 'TLE') {
				$tle ++;
			}
			elsif ($1 eq 'RED') {
				$read ++;
			}
		}
		$temp .= $_;
		if ($nesting == 0) {
			if ($read) {
				$temp =~ /^iid:(\d+)\r?\n?$/m;
				if (!$iids{$1}) {
					print($pipehandleoutput1 $temp);
				}
			}
			elsif ($tle) {
				$temp =~ /^src:(\d+)\r?\n?$/m;
				if (!$iids{$1}) {
					print($pipehandleoutput1 $temp);
				}
			}
			else {
				print($pipehandleoutput1 $temp);
			}
			undef($temp);
			undef($tle);
			undef($read);
		}
	}
	close($pipehandleoutput1);
	if ($?) {
		&errorMessage(__LINE__, "Cannot run \"$banktransact -c -f -b $bankname.temp.bnk -m -\".");
	}
	close($pipehandleinput1);
	if ($?) {
		&errorMessage(__LINE__, "Cannot run \"$bankreport -p -b $bankname\".");
	}
	if (-e "$bankname/CTG.ifo" && !-e "$bankname.temp.bnk/CTG.ifo") {
		my @ctg = glob("$bankname/CTG.*");
		foreach my $ctgfile (@ctg) {
			$ctgfile =~ /[\\\/]([^\\\/]+)$/;
			my $tempfilename = $1;
			unless (fcopy($ctgfile, "$bankname.temp.bnk/$tempfilename")) {
				&errorMessage(__LINE__, "Cannot copy \"$ctgfile\" to \"$bankname.temp.bnk/$tempfilename\".");
			}
		}
	}
	&deleteBanks($bankname);
	unless (rename("$bankname.temp.bnk", $bankname)) {
		&errorMessage(__LINE__, "Cannot rename \"$bankname.temp.bnk\" to \"$bankname\".");
	}
}

sub deleteContigsfromBank {
	my $bankname = shift(@_);
	my @temp = @_;
	my %iids;
	foreach my $iid (@temp) {
		$iids{$iid} = 1;
	}
	unless (open($pipehandleinput1, "$bankreport -p -b $bankname 2> $devnull |")) {
		&errorMessage(__LINE__, "Cannot run \"$bankreport -p -b $bankname\".");
	}
	unless (open($pipehandleoutput1, "| $banktransact -c -f -b $bankname.temp.bnk -m - 1> $devnull 2> $devnull")) {
		&errorMessage(__LINE__, "Cannot run \"$banktransact -c -f -b $bankname.temp.bnk -m -\".");
	}
	my $temp;
	my $nesting = 0;
	my $contig;
	while (<$pipehandleinput1>) {
		if ($nesting && /^\}/) {
			$nesting --;
		}
		elsif (/^\{([A-Z]+)/) {
			$nesting ++;
			if ($1 eq 'CTG') {
				$contig ++;
			}
		}
		$temp .= $_;
		if ($nesting == 0) {
			if ($contig) {
				$temp =~ /^iid:(\d+)\r?\n?$/m;
				if (!$iids{$1}) {
					print($pipehandleoutput1 $temp);
				}
			}
			else {
				print($pipehandleoutput1 $temp);
			}
			undef($temp);
			undef($contig);
		}
	}
	close($pipehandleoutput1);
	if ($?) {
		&errorMessage(__LINE__, "Cannot run \"$banktransact -c -f -b $bankname.temp.bnk -m -\".");
	}
	close($pipehandleinput1);
	if ($?) {
		&errorMessage(__LINE__, "Cannot run \"$bankreport -p -b $bankname\".");
	}
	&deleteBanks($bankname);
	unless (rename("$bankname.temp.bnk", $bankname)) {
		&errorMessage(__LINE__, "Cannot rename \"$bankname.temp.bnk\" to \"$bankname\".");
	}
}

sub dumpAllContigsSingletonsfromBank {
	my $bankname = shift(@_);
	my $filename = shift(@_);
	my $replacetype = shift(@_);
	if (-e "$bankname/CTG.ifo") {
		unless (open($pipehandleinput1, "$bank2fastq -b $bankname -iid -Q 33 2> $devnull |")) {
			&errorMessage(__LINE__, "Cannot run \"$bank2fastq -b $bankname -iid -Q 33\".");
		}
		unless (open($filehandleoutput1, ">> $filename")) {
			&errorMessage(__LINE__, "Cannot write \"$filename\".");
		}
		my $tempnline = 1;
		while (<$pipehandleinput1>) {
			if ($tempnline % 4 == 1 && /^\@/) {
				if ($replacetype) {
					s/^\@(\d+)(\r?\n?)$/\@contig_$1$2/;
				}
				else {
					s/^\@(\d+)(\r?\n?)$/\@contig_$bankname\_$1$2/;
				}
			}
			print($filehandleoutput1 $_);
			$tempnline ++;
		}
		close($filehandleoutput1);
		close($pipehandleinput1);
		if (system("$listReadPlacedStatus -S -I $bankname > $filename.singletons.iid 2> $devnull")) {
			&errorMessage(__LINE__, "Cannot run \"$listReadPlacedStatus -S -I $bankname > $filename.singletons.iid\".");
		}
		unless (open($pipehandleinput1, "$dumpreads -f -Q 33 -r -e -I $filename.singletons.iid $bankname 2> $devnull |")) {
			&errorMessage(__LINE__, "Cannot run \"$dumpreads -f -Q 33 -r -e -I $filename.singletons.iid $bankname\".");
		}
		unless (open($filehandleoutput1, ">> $filename")) {
			&errorMessage(__LINE__, "Cannot write \"$filename\".");
		}
		$tempnline = 1;
		while (<$pipehandleinput1>) {
			if ($tempnline % 4 == 1 && /^\@/ && !$replacetype) {
				s/^\@contig_(\d+)(\r?\n?)$/\@contig_$bankname\_$1$2/;
			}
			print($filehandleoutput1 $_);
			$tempnline ++;
		}
		close($filehandleoutput1);
		close($pipehandleinput1);
		unless (unlink("$filename.singletons.iid")) {
			&errorMessage(__LINE__, "Cannot delete \"$filename.singletons.iid\".");
		}
	}
	else {
		unless (open($pipehandleinput1, "$dumpreads -f -Q 33 -r -e $bankname 2> $devnull |")) {
			&errorMessage(__LINE__, "Cannot run \"$dumpreads -f -Q 33 -r -e $bankname\".");
		}
		unless (open($filehandleoutput1, ">> $filename")) {
			&errorMessage(__LINE__, "Cannot write \"$filename\".");
		}
		my $tempnline = 1;
		while (<$pipehandleinput1>) {
			if ($tempnline % 4 == 1 && /^\@/ && !$replacetype) {
				s/^\@contig_(\d+)(\r?\n?)$/\@contig_$bankname\_$1$2/;
			}
			print($filehandleoutput1 $_);
			$tempnline ++;
		}
		close($filehandleoutput1);
		close($pipehandleinput1);
	}
}

sub getContigMembersfromBank {
	my @temp = @_;
	my %contigmembers;
	my %singletons;
	foreach my $bankname (@temp) {
		my $prefix = $bankname;
		$prefix =~ s/\.bnk$//;
		if (-e "$prefix.contigmembers.gz") {
			$filehandleinput1 = &readFile("$prefix.contigmembers.gz");
			while (<$filehandleinput1>) {
				s/\r?\n?$//;
				if (my @row = split(/\t/)) {
					if (scalar(@row) > 2) {
						my $contigname = shift(@row);
						foreach my $member (@row) {
							push(@{$contigmembers{$bankname}{$contigname}}, $member);
						}
					}
					elsif (scalar(@row) == 2) {
						push(@{$singletons{$bankname}}, $row[1]);
					}
					else {
						&errorMessage(__LINE__, "Unknown error in processing a line\n$_\nof \"$prefix.contigmembers.gz\".");
					}
				}
			}
			close($filehandleinput1);
		}
		elsif (-e "$bankname/CTG.ifo") {
			# get the members
			unless (open($pipehandleinput1, "$listReadPlacedStatus $bankname 2> $devnull |")) {
				&errorMessage(__LINE__, "Cannot run \"$listReadPlacedStatus $bankname\".");
			}
			while (<$pipehandleinput1>) {
				s/\r?\n?$//;
				if (my @row = split(/\t/)) {
					if (scalar(@row) == 4 && $row[2] eq 'S') {
						push(@{$singletons{$bankname}}, $row[1]);
					}
					elsif (scalar(@row) == 5 && $row[2] eq 'P') {
						push(@{$contigmembers{$bankname}{"contig_$row[4]"}}, $row[1]);
					}
					else {
						&errorMessage(__LINE__, "Unknown error in processing a sequence \"$row[1]\" of \"contig_$row[4]\" in \"$bankname\".");
					}
				}
			}
			close($pipehandleinput1);
			if ($?) {
				&errorMessage(__LINE__, "Cannot run \"$listReadPlacedStatus $bankname\".");
			}
		}
		else {
			unless (open($pipehandleinput1, "$dumpreads -r -e $bankname 2> $devnull |")) {
				&errorMessage(__LINE__, "Cannot run \"$dumpreads -r -e $bankname\".");
			}
			while (<$pipehandleinput1>) {
				if (/^>(.+)(\r?\n?)$/) {
					push(@{$singletons{$bankname}}, $1);
				}
			}
			close($pipehandleinput1);
		}
	}
	return(\%contigmembers, \%singletons);
}

sub countAllContigsSingletonsinBank {
	my $bankname = shift(@_);
	my $tempncontigs = 0;
	if (-e "$bankname/CTG.ifo") {
		my %temp;
		unless (open($pipehandleinput1, "$listReadPlacedStatus $bankname 2> $devnull |")) {
			&errorMessage(__LINE__, "Cannot run \"$listReadPlacedStatus $bankname\".");
		}
		while (<$pipehandleinput1>) {
			if (my @row = split(/\t/)) {
				if (scalar(@row) == 4 && $row[2] eq 'S') {
					$tempncontigs ++;
				}
				elsif (scalar(@row) == 5 && $row[2] eq 'P' && !$temp{$row[4]}) {
					$temp{$row[4]} = 1;
					$tempncontigs ++;
				}
			}
		}
		close($pipehandleinput1);
		if ($?) {
			&errorMessage(__LINE__, "Cannot run \"$listReadPlacedStatus $bankname\".");
		}
	}
	else {
		unless (open($pipehandleinput1, "$dumpreads -r $bankname 2> $devnull |")) {
			&errorMessage(__LINE__, "Cannot run \"$dumpreads -r $bankname\".");
		}
		while (<$pipehandleinput1>) {
			if (/^>/) {
				$tempncontigs ++;
			}
		}
		close($pipehandleinput1);
		if ($?) {
			&errorMessage(__LINE__, "Cannot run \"$dumpreads -r $bankname\".");
		}
	}
	if ($tempncontigs == 0) {
		&errorMessage(__LINE__, "Cannot find contigs and singletons in \"$bankname\".");
	}
	return($tempncontigs);
}

sub combineBanks {
	my $bankname = pop(@_);
	my @banks = @_;
	my $tempbank = shift(@banks);
	if (@banks) {
		while (my @tempbanks = splice(@banks, 0, 19)) {
			if (system("$bankcombine $bankname $tempbank @tempbanks 1> $devnull 2> $devnull")) {
				&errorMessage(__LINE__, "Cannot run \"$bankcombine $bankname $tempbank @tempbanks\".");
			}
			&deleteBanks($tempbank);
			if (!$nodel) {
				&deleteBanks(@tempbanks);
			}
			if (@banks) {
				unless (rename($bankname, $tempbank)) {
					&errorMessage(__LINE__, "Cannot rename \"$bankname\" to \"$tempbank\".");
				}
			}
		}
	}
	else {
		unless (rename($tempbank, $bankname)) {
			&errorMessage(__LINE__, "Cannot rename \"$tempbank\" to \"$bankname\".");
		}
	}
}

sub deleteSingletonContigs {
	my $bankname = shift(@_);
	unless (open($pipehandleinput1, "$bankreport -p -b $bankname 2> $devnull |")) {
		&errorMessage(__LINE__, "Cannot run \"$bankreport -p -b $bankname\".");
	}
	unless (open($pipehandleoutput1, "| $banktransact -c -f -b $bankname.temp.bnk -m - 1> $devnull 2> $devnull")) {
		&errorMessage(__LINE__, "Cannot run \"$banktransact -c -f -b $bankname.temp.bnk -m -\".");
	}
	my $temp;
	my $nesting = 0;
	my $contig;
	while (<$pipehandleinput1>) {
		if ($nesting && /^\}/) {
			$nesting --;
		}
		elsif (/^\{([A-Z]+)/) {
			$nesting ++;
			if ($1 eq 'CTG') {
				$contig ++;
			}
		}
		$temp .= $_;
		if ($nesting == 0) {
			if ($contig) {
				my @tle = $temp =~ /^\{TLE/gm;
				if (scalar(@tle) > 1) {
					print($pipehandleoutput1 $temp);
				}
			}
			else {
				print($pipehandleoutput1 $temp);
			}
			undef($temp);
			undef($contig);
		}
	}
	close($pipehandleoutput1);
	if ($?) {
		&errorMessage(__LINE__, "Cannot run \"$banktransact -c -f -b $bankname.temp.bnk -m -\".");
	}
	close($pipehandleinput1);
	if ($?) {
		&errorMessage(__LINE__, "Cannot run \"$bankreport -p -b $bankname\".");
	}
	if (!-e "$bankname.temp.bnk/CTG.ifo" && -e "$bankname/CTG.ifo") {
		my @ctg = glob("$bankname/CTG.*");
		foreach my $ctgfile (@ctg) {
			$ctgfile =~ /[\\\/]([^\\\/]+)$/;
			my $tempfilename = $1;
			unless (fcopy($ctgfile, "$bankname.temp.bnk/$tempfilename")) {
				&errorMessage(__LINE__, "Cannot copy \"$ctgfile\" to \"$bankname.temp.bnk/$tempfilename\".");
			}
		}
	}
	&deleteBanks($bankname);
	unless (rename("$bankname.temp.bnk", $bankname)) {
		&errorMessage(__LINE__, "Cannot rename \"$bankname.temp.bnk\" to \"$bankname\".");
	}
}

sub deleteBanks {
	foreach my $bankname (@_) {
		while (glob("$bankname/*")) {
			unless (unlink($_)) {
				&errorMessage(__LINE__, "Cannot delete \"$_\".");
			}
		}
		if (-e $bankname && -d $bankname) {
			unless (rmdir($bankname)) {
				&errorMessage(__LINE__, "Cannot delete \"$bankname\".");
			}
		}
	}
}

sub deleteContigMembers {
	foreach my $bankname (@_) {
		my $contigmembers = $bankname;
		$contigmembers =~ s/\.bnk$/.contigmembers.gz/;
		if (-e $contigmembers) {
			unless (unlink($contigmembers)) {
				&errorMessage(__LINE__, "Cannot delete \"$contigmembers\".");
			}
		}
	}
}

sub cleanBank {
	my $bankname = shift(@_);
	my $keep = join('|', @_);
	while (glob("$bankname/*")) {
		if ($_ !~ /(?:$keep)\.[0-9a-z\.]+$/ && !unlink($_)) {
			&errorMessage(__LINE__, "Cannot delete \"$_\".");
		}
	}
}

# error message
sub errorMessage {
	my $lineno = shift(@_);
	my $message = shift(@_);
	print(STDERR "ERROR!: line $lineno\n$message\n");
	print(STDERR "If you want to read help message, run this script without options.\n");
	exit(1);
}

sub helpMessage {
	print(STDERR <<"_END");
Usage
=====
assams options input output

Command line options
====================
--assemblemode=BLAT|BLATFAST|MINIMUS
  Specify assemble mode. (default: MINIMUS)

--mappingmode=BLAT|BLATFAST|MINIMUS|NONE
  Specify assemble mode. (default: MINIMUS)

--linkagemode=SINGLE|COMPLETE
  Specify linkage mode. (default: COMPLETE)

--mergemode=ROUGH|DRAFT|NORMAL|ACCURATE
  Specify merging mode. (default: NORMAL)

--minident=DECIMAL(,DECIMAL,DECIMAL)
  Specify minimum identity percentage. First value will be used in primary
assembly, second one will be used in secondary assembly, and last one
will be used in mapping. 0 means automatic. (default: 0.97,0.97,0.97)

--minovllen=INTEGER
  Specify minimum overlap length. 0 means automatic. (default: 100)

--strand=PLUS|BOTH
  Specify search strand. (default: BOTH)

--alnwiggle=INTEGER
  Specify alignment wiggle. (default: 11)

--minnseq=INTEGER
  Specify minimum number of sequences of each splitted file.
(default: 1000)

--maxnsplit=INTEGER
  Specify maximum number of splitting. (default: unlimited)

--addniter=INTEGER
  Specify the number of additional iteration. (default: 0)

--numthreads=INTEGER
  Specify the number of threads. (default: 1)

Acceptable input file formats
=============================
FASTQ (uncompressed, gzip-compressed, bzip2-compressed, or xz-compressed)
(Quality values must be encoded in Sanger format.)
AMOS bank

Output format
=============
AMOS bank
_END
	exit;
}
